import logging
from rich.logging import RichHandler

class ColorScheme:
    domain = "[green]"
    user = "[sea_green3]"
    computer = "[red1]"
    group = "[gold1]"
    enterpriseca = "[medium_purple1]"
    aiaca = "[medium_purple2]"
    rootca = "[medium_purple3]"
    certtemplate = "[bright_magenta]"
    schema = "[deep_sky_blue1]"
    ou = "[dark_orange]"
    containers = "[orange]"
    gpo = "[purple]"

OBJ_EXTRA_FMT = {
    "markup": True,
    "highlighter": False
}

FORMAT = "%(message)s"
logger = logging.getLogger("bofhound")

handler = RichHandler(omit_repeated_times=False, show_path=False, keywords=[])
handler.setFormatter(logging.Formatter(FORMAT, datefmt="[%X]"))
logger.addHandler(handler)

logger.propagate = False



import ipaddress

from bofhound.logger import logger, ColorScheme, OBJ_EXTRA_FMT


class LocalRegistrySession:
    REGSESSION_HOST     = "Host"
    REGSESSION_USER_SID = "UserSid"


    def __init__(self, object):
        self.user_sid       = None
        self.host_name      = None
        self.host_domain    = None
        self.host_fqdn      = None

        try:
            ipaddress.ip_address(object[LocalRegistrySession.REGSESSION_HOST])
            logger.debug(f"Skipping session on {object[LocalRegistrySession.REGSESSION_HOST]} due to IP instead of hostname")
            return
        except:
            pass
        
        # will be set to True if correlated to a computer object
        self.matched = False

        if LocalRegistrySession.REGSESSION_USER_SID in object.keys():
            self.user_sid = object[LocalRegistrySession.REGSESSION_USER_SID]

        if LocalRegistrySession.REGSESSION_HOST in object.keys():
            if '.' in object[LocalRegistrySession.REGSESSION_HOST]:
                self.host_fqdn = object[LocalRegistrySession.REGSESSION_HOST]
                self.host_name = self.host_fqdn.split('.')[0]
                self.host_domain = '.'.join(self.host_fqdn.split('.')[1:])
            else:
                self.host_name = object[LocalRegistrySession.REGSESSION_HOST]
                logger.debug(f"FQDN missing from hostname for {ColorScheme.user}{self.user_sid}[/] session on {ColorScheme.computer}{self.host_name}[/]", extra=OBJ_EXTRA_FMT)

    def should_import(self, known_domain_sids):
        # missing required attributes
        if self.user_sid is None or self.host_name is None:
            return False

        # do not import local account sessions or 
        # user sessions from unknown domains
        if self.user_sid.rsplit('-', 1)[0] not in known_domain_sids:
            logger.debug(f"Skipping session for {ColorScheme.user}{self.user_sid}[/] since domain SID is unfamiliar", extra=OBJ_EXTRA_FMT)
            return False

        computer = self.host_fqdn if self.host_fqdn else self.host_name
        logger.debug(f"Registry session found for {ColorScheme.user}{self.user_sid}[/] on {ColorScheme.computer}{computer}[/]", extra=OBJ_EXTRA_FMT)
        return True


    # so that a set can be used to keep a unique list of objects
    def __eq__(self, other):
        return (self.user_sid, self.host_name, self.host_domain) == \
               (other.user_sid, other.host_name, other.host_domain)
    

    # so that a set can be used to keep a unique list of objects
    def __hash__(self):
        return hash((self.user_sid, self.host_name, self.host_domain))


    # for debugging
    def __repr__(self):
        return f"RegistrySession<user_sid={self.user_sid}, host_name={self.host_name}, host_domain={self.host_domain}>"


from bloodhound.ad.utils import ADUtils
from .bloodhound_object import BloodHoundObject
import ast
import base64

from bofhound.ad.helpers.cert_utils import OID_TO_STR_MAP, MS_PKI_CERTIFICATE_NAME_FLAG, MS_PKI_PRIVATE_KEY_FLAG, MS_PKI_ENROLLMENT_FLAG, filetime_to_span, span_to_str
from bofhound.logger import logger, OBJ_EXTRA_FMT, ColorScheme


class BloodHoundCertTemplate(BloodHoundObject):
    GUI_PROPERTIES = [
        'domain', 'name', 'distinguishedname', 'domainsid', 'isaclprotected',
        'description', 'whencreated', 'validityperiod', 'renewalperiod',
        'schemaversion', 'displayname', 'oid', 'enrollmentflag', 'requiresmanagerapproval',
        'nosecurityextension', 'certificatenameflag', 'enrolleesuppliessubject', 'subjectaltrequireupn',
        'subjectaltrequiredns', 'subjectaltrequiredomaindns', 'subjectaltrequireemail',
        'subjectaltrequirespn', 'subjectrequireemail', 'ekus', 'certificateapplicationpolicy',
        'authorizedsignatures', 'applicationpolicies', 'issuancepolicies', 'effectiveekus',
        'authenticationenabled'
    ]

    COMMON_PROPERTIES = [
    ]

    def __init__(self, object):

        super().__init__(object)

        self._entry_type = "PKI Template"
        self.GPLinks = []
        self.ContainedBy = {}
        self.IsACLProtected = False
        self.cas_ids = []

        if 'objectguid' in object.keys():
            self.ObjectIdentifier = object.get("objectguid").upper()

        if 'distinguishedname' in object.keys():
            domain = ADUtils.ldap2domain(object.get('distinguishedname')).upper()
            self.Properties['domain'] = domain
            self.Properties['distinguishedname'] = object.get('distinguishedname').upper()
            self.Properties['name'] = self.get_cn_from_dn(object.get('distinguishedname')) + "@" + domain

        if 'description' in object.keys():
            self.Properties['description'] = object.get('description')
        else:
            self.Properties['description'] = None

        if 'pkiexpirationperiod' in object.keys():
            pKIExpirationPeriod_b64 = object.get("pkiexpirationperiod")
            pKIExpirationPeriod_byte_array = base64.b64decode(pKIExpirationPeriod_b64)
            self.Properties["validityperiod"] = span_to_str(filetime_to_span(pKIExpirationPeriod_byte_array))

        if 'pkioverlapperiod' in object.keys():
            pKIRenewalPeriod_b64 = object.get("pkioverlapperiod")
            pKIRenewalPeriod_byte_array = base64.b64decode(pKIRenewalPeriod_b64)
            self.Properties["renewalperiod"] = span_to_str(filetime_to_span(pKIRenewalPeriod_byte_array))

        if 'mspki-template-schema-version' in object.keys():
            self.Properties['schemaversion'] = int(object.get('mspki-template-schema-version'))

        if 'displayname' in object.keys():
            self.Properties["displayname"] = object.get("displayname")

        if 'mspki-cert-template-oid' in object.keys():
            self.Properties["oid"] = object.get("mspki-cert-template-oid")

        if 'mspki-enrollment-flag' in object.keys():
            enrollment_flag = object.get("mspki-enrollment-flag")
            if enrollment_flag is not None:
                enrollment_flag = MS_PKI_ENROLLMENT_FLAG(int(enrollment_flag))
            else:
                enrollment_flag = MS_PKI_ENROLLMENT_FLAG(0)
            self.Properties["enrollmentflag"] = ', '.join(enrollment_flag.to_str_list())

            requires_manager_approval = (
                MS_PKI_ENROLLMENT_FLAG.PEND_ALL_REQUESTS in enrollment_flag
            )
            self.Properties["requiresmanagerapproval"] = requires_manager_approval

            no_security_extension = (
                MS_PKI_ENROLLMENT_FLAG.NO_SECURITY_EXTENSION in enrollment_flag
            )
            self.Properties["nosecurityextension"] = no_security_extension
        
        if 'mspki-certificate-name-flag' in object.keys():
            certificate_name_flag = object.get('mspki-certificate-name-flag')
            if certificate_name_flag is not None:
                certificate_name_flag = MS_PKI_CERTIFICATE_NAME_FLAG(
                    int(certificate_name_flag)
                )
            else:
                certificate_name_flag = MS_PKI_CERTIFICATE_NAME_FLAG(0)
            self.Properties["certificatenameflag"] = ", ".join(certificate_name_flag.to_str_list())

            enrollee_supplies_subject = any(
                flag in certificate_name_flag
                for flag in [
                    MS_PKI_CERTIFICATE_NAME_FLAG.ENROLLEE_SUPPLIES_SUBJECT,
                ]
            )
            self.Properties["enrolleesuppliessubject"] = enrollee_supplies_subject

            subjectaltrequireupn = any(
                flag in certificate_name_flag
                for flag in [
                    MS_PKI_CERTIFICATE_NAME_FLAG.SUBJECT_ALT_REQUIRE_UPN,
                ]
            )
            self.Properties["subjectaltrequireupn"] = subjectaltrequireupn

            subjectaltrequiredns = any(
                flag in certificate_name_flag
                for flag in [
                    MS_PKI_CERTIFICATE_NAME_FLAG.SUBJECT_ALT_REQUIRE_DNS,
                ]
            )
            self.Properties["subjectaltrequiredns"] = subjectaltrequiredns

            subjectaltrequiredomaindns = any(
                flag in certificate_name_flag
                for flag in [
                    MS_PKI_CERTIFICATE_NAME_FLAG.SUBJECT_ALT_REQUIRE_DOMAIN_DNS,
                ]
            )
            self.Properties["subjectaltrequiredomaindns"] = subjectaltrequiredomaindns

            subjectaltrequireemail = any(
                flag in certificate_name_flag
                for flag in [
                    MS_PKI_CERTIFICATE_NAME_FLAG.SUBJECT_ALT_REQUIRE_EMAIL,
                ]
            )
            self.Properties["subjectaltrequireemail"] = subjectaltrequireemail

            subjectaltrequirespn = any(
                flag in certificate_name_flag
                for flag in [
                    MS_PKI_CERTIFICATE_NAME_FLAG.SUBJECT_ALT_REQUIRE_SPN,
                ]
            )
            self.Properties["subjectaltrequirespn"] = subjectaltrequirespn

            subjectrequireemail = any(
                flag in certificate_name_flag
                for flag in [
                    MS_PKI_CERTIFICATE_NAME_FLAG.SUBJECT_REQUIRE_EMAIL,
                ]
            )
            self.Properties["subjectrequireemail"] = subjectrequireemail

        ekus = []
        if 'pkiextendedkeyusage' in object.keys():
            ekus = object.get('pkiextendedkeyusage').split(', ')

            ### This parses ekus to readable format
            #ekus = list(
            #    map(lambda x: OID_TO_STR_MAP[x] if x in OID_TO_STR_MAP else x, ekus)
            #)
            self.Properties["ekus"] = ekus

            ### Not needed anymore
            #any_purpose = (
            #    "Any Purpose" in ekus or len(ekus) == 0
            #)
            #client_authentication = any_purpose or any(
            #    eku in ekus
            #    for eku in [
            #        "Client Authentication",
            #        "Smart Card Logon",
            #        "PKINIT Client Authentication",
            #    ]
            #)
            #enrollment_agent = any_purpose or any(
            #    eku in ekus
            #    for eku in [
            #        "Certificate Request Agent",
            #    ]
            #)

            #self.Properties["Client Authentication"] = client_authentication
            #self.Properties["Enrollment Agent"] = enrollment_agent            
            #self.Properties["Any Purpose"] = any_purpose
        
        #self.Properties["Certificate Authorities"] = []
        #self.Properties["Enabled"] = True

        if 'mspki-certificate-application-policy' in object.keys():
            self.Properties['certificateapplicationpolicy'] = object.get('mspki-certificate-application-policy').split(', ')
        else:
            self.Properties['certificateapplicationpolicy'] = []
            
        if 'mspki-ra-signature' in object.keys():
            authorized_signatures_required = object.get("mspki-ra-signature")
            if authorized_signatures_required is not None:
                authorized_signatures_required = int(authorized_signatures_required)
            else:
                authorized_signatures_required = 0
            self.Properties["authorizedsignatures"] = authorized_signatures_required
            


        self.Properties['applicationpolicies'] = []
        if 'mspki-private-key-flag' in object.keys():
            private_key_flag = object.get("mspki-private-key-flag")
            if private_key_flag is not None:
                private_key_flag = MS_PKI_PRIVATE_KEY_FLAG(int(private_key_flag))
            else:
                private_key_flag = MS_PKI_PRIVATE_KEY_FLAG(0)

        if 'mspki-ra-application-policies' in object.keys():
            applicationpolicies = object.get('mspki-ra-application-policies')
            schemaversions_noparsing = [0, 1, 2]

            hasUseLegacyProvider = (
                MS_PKI_PRIVATE_KEY_FLAG.USE_LEGACY_PROVIDER in private_key_flag
            )

            if (self.Properties['schemaversion'] in schemaversions_noparsing or
            (self.Properties['schemaversion'] == 4 and hasUseLegacyProvider)):
                self.Properties['applicationpolicies'] = applicationpolicies
            
            else:
                # Divide entries into groups of three, filter, and select third element
                result = [parts[2] for parts in (applicationpolicies[i*3:i*3+3] for i in range(len(applicationpolicies)//3))
                        if len(parts) == 3 and parts[0] == 'mspki-ra-application-policies']

                # Convert the result to an array
                result_array = list(result)
                self.Properties['applicationpolicies'] = result_array

        if 'mspki-ra-policies' in object.keys():
            self.Properties['issuancepolicies'] = object.get('mspki-ra-policies').split(', ')
        else:
            self.Properties['issuancepolicies'] = []

        if (self.Properties['schemaversion'] == 1 and len(ekus)>0):
            self.Properties['effectiveekus'] = ekus
        elif self.Properties['certificateapplicationpolicy']:
            self.Properties['effectiveekus'] = self.Properties['certificateapplicationpolicy']
        else:
            self.Properties['effectiveekus'] = []
        
        AuthenticationOIDs = ['1.3.6.1.5.5.7.3.2', '1.3.6.1.5.2.3.4', '1.3.6.1.4.1.311.20.2.2', '2.5.29.37.0']
        self.Properties['authenticationenabled'] = bool(set(self.Properties['effectiveekus']).intersection(AuthenticationOIDs)) or len(self.Properties['effectiveekus']) == 0

        if 'ntsecuritydescriptor' in object.keys():
            self.RawAces = object['ntsecuritydescriptor']

    def to_json(self, properties_level):
        self.Properties['isaclprotected'] = self.IsACLProtected
        data = super().to_json(properties_level)
        data["ObjectIdentifier"] = self.ObjectIdentifier
        data["ContainedBy"] = self.ContainedBy
        data["Aces"] = self.Aces
        data['IsACLProtected'] = self.IsACLProtected
        data["cas_ids"] = self.cas_ids
        return data


from bloodhound.ad.utils import ADUtils
from .bloodhound_object import BloodHoundObject

from bofhound.logger import logger, OBJ_EXTRA_FMT, ColorScheme


class BloodHoundDomain(BloodHoundObject):

    GUI_PROPERTIES = [
        'distinguishedname', 'domainsid', 'description', 'whencreated',
        'functionallevel', 'domain', 'isaclprotected', 'collected',
        'name'
    ]

    COMMON_PROPERTIES = [
    ]

    def __init__(self, object):
        super().__init__(object)

        self._entry_type = "Domain"
        self.GPLinks = []
        self.ContainedBy = {}
        level_id = object.get('msds-behavior-version', 0)
        try:
            functional_level = ADUtils.FUNCTIONAL_LEVELS[int(level_id)]
        except KeyError:
            functional_level = 'Unknown'

        dc = None

        self.Properties['collected'] = True

        if 'distinguishedname' in object.keys():
            self.Properties["name"] = ADUtils.ldap2domain(object.get('distinguishedname').upper())
            self.Properties["domain"] = self.Properties["name"]
            dc = BloodHoundObject.get_domain_component(object.get('distinguishedname').upper())
            logger.debug(f"Reading Domain object {ColorScheme.domain}{self.Properties['name']}[/]", extra=OBJ_EXTRA_FMT)

        if 'objectsid' in object.keys():
            self.Properties["domainsid"] = object.get('objectsid')

        if 'distinguishedname' in object.keys():
            self.Properties['distinguishedname'] = object.get('distinguishedname').upper()

        if 'description' in object.keys():
            self.Properties["description"] = object.get('description')
        else:
            self.Properties["description"] = None

        if 'ntsecuritydescriptor' in object.keys():
            self.RawAces = object['ntsecuritydescriptor']
        
        if 'gplink' in object.keys():
            # this is gross - not sure why gplink is coming in without a colon
            # (even from logs in test folder) but will hunt down later if it's a problem
            links = object.get('gplink').replace('LDAP//', 'LDAP://')
            
            # [['DN1', 'GPLinkOptions1'], ['DN2', 'GPLinkOptions2'], ...]
            self.GPLinks = [link.upper()[:-1].split(';') for link in links.split('[LDAP://')][1:]

        self.Properties["highvalue"] = True

        self.Properties["functionallevel"] = functional_level

        self.Trusts = []
        self.Aces = []
        self.Links = []
        self.ChildObjects = []
        self.GPOChanges = {
            "AffectedComputers": [],
            "DcomUsers": [],
            "LocalAdmins": [],
            "PSRemoteUsers": [],
            "RemoteDesktopUsers": []
        }
        self.IsDeleted = False
        self.IsACLProtected = False


    def to_json(self, properties_level):
        self.Properties['isaclprotected'] = self.IsACLProtected
        domain = super().to_json(properties_level)

        domain["ObjectIdentifier"] = self.ObjectIdentifier
        domain["Trusts"] = self.Trusts
        domain["ContainedBy"] = None
        # The below is all unsupported as of now.
        domain["Aces"] = self.Aces
        domain["Links"] = self.Links
        domain["ChildObjects"] = self.ChildObjects
        
        self.GPOChanges["AffectedComputers"] = self.AffectedComputers
        self.GPOChanges["AffectedUsers"] = self.AffectedUsers
        domain["GPOChanges"] = self.GPOChanges

        domain["IsDeleted"] = self.IsDeleted
        domain["IsACLProtected"] = self.IsACLProtected

        return domain




class SharedParser():

    def __init__(self):
        pass


    # will be same for all child classes
    @staticmethod
    def parse_line(line, current_object):
        data = line.split(': ')
        try:
            attr = data[0].strip()
            value = data[1].strip()

            # if attr is not on the current object, add it
            if attr not in current_object:
                current_object[attr] = value

            return current_object
        except IndexError:
            return current_object


    # will be implemented in child classes
    @staticmethod
    def is_start_boundary_line(line):
        pass
    

    # will be implemented in child classes
    @staticmethod
    def is_end_boundary_line(line):
        pass


import ipaddress

from bofhound.logger import logger, ColorScheme, OBJ_EXTRA_FMT


class LocalGroupMembership:
    LOCALGROUP_HOST              = "Host"
    LOCALGROUP_GROUP             = "Group"
    LOCALGROUP_MEMBER            = "Member"
    LOCALGROUP_MEMBER_SID        = "MemberSid"
    LOCALGROUP_MEMBER_SID_TYPE   = "MemberSidType"
    LOCALGROUP_NAMES             = [
        "administrators",
        "remote desktop users",
        "remote management users",
        "distributed com users"
    ]

    def __init__(self, object):
        self.host_name              = None
        self.host_domain            = None
        self.host_fqdn              = None
        self.group                  = None
        self.member                 = None
        self.member_netbios_domain  = None
        self.member_sid             = None
        self.member_sid_type        = None

        # will be set to True if correlated to a computer object
        self.matched = False

        try:
            ipaddress.ip_address(object[LocalGroupMembership.LOCALGROUP_HOST])
            logger.debug(f"Skipping local group member on {object[LocalGroupMembership.LOCALGROUP_HOST]} due to IP instead of hostname")
            return
        except:
            pass

        if LocalGroupMembership.LOCALGROUP_GROUP in object.keys():
            self.group = object[LocalGroupMembership.LOCALGROUP_GROUP]

        if LocalGroupMembership.LOCALGROUP_MEMBER in object.keys():
            parts = object[LocalGroupMembership.LOCALGROUP_MEMBER].split('\\')
            if len(parts) == 2:
                self.member = parts[1]
                self.member_netbios_domain = parts[0]
            else:
                self.member = parts[0]

        if LocalGroupMembership.LOCALGROUP_HOST in object.keys():
            if '.' in object[LocalGroupMembership.LOCALGROUP_HOST]:
                self.host_name = object[LocalGroupMembership.LOCALGROUP_HOST].split('.')[0]
                self.host_domain = '.'.join(object[LocalGroupMembership.LOCALGROUP_HOST].split('.')[1:])
                self.host_fqdn = object[LocalGroupMembership.LOCALGROUP_HOST]
            else:
                self.host_name = object[LocalGroupMembership.LOCALGROUP_HOST]
                logger.debug(f"FQDN missing from hostname for {ColorScheme.user}{self.user}[/] session on {ColorScheme.computer}{self.host_name}[/]", extra=OBJ_EXTRA_FMT)

        if LocalGroupMembership.LOCALGROUP_MEMBER_SID in object.keys():
            self.member_sid = object[LocalGroupMembership.LOCALGROUP_MEMBER_SID]

        if LocalGroupMembership.LOCALGROUP_MEMBER_SID_TYPE in object.keys():
            self.member_sid_type = object[LocalGroupMembership.LOCALGROUP_MEMBER_SID_TYPE]
    

    def should_import(self, known_domain_sids):
        # missing required attributes
        if self.host_name is None or self.group is None \
            or self.member_sid is None or self.member_sid_type is None:
            return False
        
        # filter out local groups we don't care about
        if self.group.lower() not in LocalGroupMembership.LOCALGROUP_NAMES:
            return False
        
        # do not import local account sessions or 
        # user sessions from unknown domains
        if self.member_sid.rsplit('-', 1)[0] not in known_domain_sids:
            color = ColorScheme.user if self.member_sid_type == "User" else ColorScheme.group
            logger.debug(f"Skipping local group membership for {color}{self.member}[/] since domain SID is unfamiliar", extra=OBJ_EXTRA_FMT)
            return False
        
        computer = f"{self.host_name}.{self.host_domain}" if self.host_domain else self.host_name
        user = f"{self.member}@{self.member_netbios_domain}" if self.member_netbios_domain else self.member
        logger.debug(f"Local group member found for {ColorScheme.user}{user}[/] on {ColorScheme.computer}{computer}[/]", extra=OBJ_EXTRA_FMT)
        return True
    

    # so that a set can be used to keep a unique list of objects
    def __eq__(self, other):
        return (self.host_name, self.host_domain, self.group, self.member_sid) == \
               (other.host_name, other.host_domain, other.group, other.member_sid)
    

    # so that a set can be used to keep a unique list of objects
    def __hash__(self):
        return hash((self.host_name, self.host_domain, self.group, self.member_sid))
    

    # for debugging
    def __repr__(self):
        return f"LocalGroupMembership(host_name={self.host_name}, group={self.group}, member={self.member}, member_netbios_domain={self.member_netbios_domain}, member_sid={self.member_sid}, member_sid_type={self.member_sid_type})"
    



from .sharedparser import SharedParser


class NetLocalGroupBofParser(SharedParser):
    START_BOUNDARY  = "----------Local Group Member----------"
    END_BOUNDARY    = "--------End Local Group Member--------"
    OBJECT_TYPE     = "LocalGroup"

    def __init__(self):
        pass


    @staticmethod
    def is_start_boundary_line(line):
        return line.strip() == NetLocalGroupBofParser.START_BOUNDARY
    

    @staticmethod
    def is_end_boundary_line(line):
        return line.strip() == NetLocalGroupBofParser.END_BOUNDARY


## From Certipy https://github.com/ly4k/Certipy/blob/main/certipy/lib/constants.py
##              https://github.com/ly4k/Certipy/blob/main/certipy/lib/structs.py
##              https://github.com/ly4k/Certipy/blob/main/certipy/lib/formatting.py

import enum
import struct

def filetime_to_span(filetime: str) -> int:
    (span,) = struct.unpack("<q", filetime)

    span *= -0.0000001

    return int(span)

def span_to_str(span: int) -> str:
    if (span % 31536000 == 0) and (span // 31536000) >= 1:
        if (span / 31536000) == 1:
            return "1 year"
        return "%i years" % (span // 31536000)
    elif (span % 2592000 == 0) and (span // 2592000) >= 1:
        if (span // 2592000) == 1:
            return "1 month"
        else:
            return "%i months" % (span // 2592000)
    elif (span % 604800 == 0) and (span // 604800) >= 1:
        if (span / 604800) == 1:
            return "1 week"
        else:
            return "%i weeks" % (span // 604800)

    elif (span % 86400 == 0) and (span // 86400) >= 1:
        if (span // 86400) == 1:
            return "1 day"
        else:
            return "%i days" % (span // 86400)
    elif (span % 3600 == 0) and (span / 3600) >= 1:
        if (span // 3600) == 1:
            return "1 hour"
        else:
            return "%i hours" % (span // 3600)
    else:
        return ""

def to_pascal_case(snake_str: str) -> str:
    components = snake_str.split("_")
    return "".join(x.title() for x in components)

def _high_bit(value):
    """returns index of highest bit, or -1 if value is zero or negative"""
    return value.bit_length() - 1

def _decompose(flag, value):
    """Extract all members from the value."""
    # _decompose is only called if the value is not named
    not_covered = value
    negative = value < 0
    members = []
    for member in flag:
        member_value = member.value
        if member_value and member_value & value == member_value:
            members.append(member)
            not_covered &= ~member_value
    if not negative:
        tmp = not_covered
        while tmp:
            flag_value = 2 ** _high_bit(tmp)
            if flag_value in flag._value2member_map_:
                members.append(flag._value2member_map_[flag_value])
                not_covered &= ~flag_value
            tmp &= ~flag_value
    if not members and value in flag._value2member_map_:
        members.append(flag._value2member_map_[value])
    members.sort(key=lambda m: m._value_, reverse=True)
    if len(members) > 1 and members[0].value == value:
        # we have the breakdown, don't need the value member itself
        members.pop(0)
    return members, not_covered

class PkiCertificateAuthorityFlags(enum.Enum):
    NO_TEMPLATE_SUPPORT = 1
    SUPPORTS_NT_AUTHENTICATION = 2
    CA_SUPPORTS_MANUAL_AUTHENTICATION = 4
    CA_SERVERTYPE_ADVANCED = 8

OID_TO_STR_MAP = {
    "1.3.6.1.4.1.311.76.6.1": "Windows Update",
    "1.3.6.1.4.1.311.10.3.11": "Key Recovery",
    "1.3.6.1.4.1.311.10.3.25": "Windows Third Party Application Component",
    "1.3.6.1.4.1.311.21.6": "Key Recovery Agent",
    "1.3.6.1.4.1.311.10.3.6": "Windows System Component Verification",
    "1.3.6.1.4.1.311.61.4.1": "Early Launch Antimalware Drive",
    "1.3.6.1.4.1.311.10.3.23": "Windows TCB Component",
    "1.3.6.1.4.1.311.61.1.1": "Kernel Mode Code Signing",
    "1.3.6.1.4.1.311.10.3.26": "Windows Software Extension Verification",
    "2.23.133.8.3": "Attestation Identity Key Certificate",
    "1.3.6.1.4.1.311.76.3.1": "Windows Store",
    "1.3.6.1.4.1.311.10.6.1": "Key Pack Licenses",
    "1.3.6.1.4.1.311.20.2.2": "Smart Card Logon",
    "1.3.6.1.5.2.3.5": "KDC Authentication",
    "1.3.6.1.5.5.7.3.7": "IP security use",
    "1.3.6.1.4.1.311.10.3.8": "Embedded Windows System Component Verification",
    "1.3.6.1.4.1.311.10.3.20": "Windows Kits Component",
    "1.3.6.1.5.5.7.3.6": "IP security tunnel termination",
    "1.3.6.1.4.1.311.10.3.5": "Windows Hardware Driver Verification",
    "1.3.6.1.5.5.8.2.2": "IP security IKE intermediate",
    "1.3.6.1.4.1.311.10.3.39": "Windows Hardware Driver Extended Verification",
    "1.3.6.1.4.1.311.10.6.2": "License Server Verification",
    "1.3.6.1.4.1.311.10.3.5.1": "Windows Hardware Driver Attested Verification",
    "1.3.6.1.4.1.311.76.5.1": "Dynamic Code Generato",
    "1.3.6.1.5.5.7.3.8": "Time Stamping",
    "1.3.6.1.4.1.311.10.3.4.1": "File Recovery",
    "1.3.6.1.4.1.311.2.6.1": "SpcRelaxedPEMarkerCheck",
    "2.23.133.8.1": "Endorsement Key Certificate",
    "1.3.6.1.4.1.311.2.6.2": "SpcEncryptedDigestRetryCount",
    "1.3.6.1.4.1.311.10.3.4": "Encrypting File System",
    "1.3.6.1.5.5.7.3.1": "Server Authentication",
    "1.3.6.1.4.1.311.61.5.1": "HAL Extension",
    "1.3.6.1.5.5.7.3.4": "Secure Email",
    "1.3.6.1.5.5.7.3.5": "IP security end system",
    "1.3.6.1.4.1.311.10.3.9": "Root List Signe",
    "1.3.6.1.4.1.311.10.3.30": "Disallowed List",
    "1.3.6.1.4.1.311.10.3.19": "Revoked List Signe",
    "1.3.6.1.4.1.311.10.3.21": "Windows RT Verification",
    "1.3.6.1.4.1.311.10.3.10": "Qualified Subordination",
    "1.3.6.1.4.1.311.10.3.12": "Document Signing",
    "1.3.6.1.4.1.311.10.3.24": "Protected Process Verification",
    "1.3.6.1.4.1.311.80.1": "Document Encryption",
    "1.3.6.1.4.1.311.10.3.22": "Protected Process Light Verification",
    "1.3.6.1.4.1.311.21.19": "Directory Service Email Replication",
    "1.3.6.1.4.1.311.21.5": "Private Key Archival",
    "1.3.6.1.4.1.311.10.5.1": "Digital Rights",
    "1.3.6.1.4.1.311.10.3.27": "Preview Build Signing",
    "1.3.6.1.4.1.311.20.2.1": "Certificate Request Agent",
    "2.23.133.8.2": "Platform Certificate",
    "1.3.6.1.4.1.311.20.1": "CTL Usage",
    "1.3.6.1.5.5.7.3.9": "OCSP Signing",
    "1.3.6.1.5.5.7.3.3": "Code Signing",
    "1.3.6.1.4.1.311.10.3.1": "Microsoft Trust List Signing",
    "1.3.6.1.4.1.311.10.3.2": "Microsoft Time Stamping",
    "1.3.6.1.4.1.311.76.8.1": "Microsoft Publishe",
    "1.3.6.1.5.5.7.3.2": "Client Authentication",
    "1.3.6.1.5.2.3.4": "PKINIT Client Authentication",
    "1.3.6.1.4.1.311.10.3.13": "Lifetime Signing",
    "2.5.29.37.0": "Any Purpose",
    "1.3.6.1.4.1.311.64.1.1": "Server Trust",
    "1.3.6.1.4.1.311.10.3.7": "OEM Windows System Component Verification",
}

class IntFlag(enum.IntFlag):
    def to_list(self):
        cls = self.__class__
        members, _ = _decompose(cls, self._value_)
        return members

    def to_str_list(self):
        return list(map(lambda x: str(x), self.to_list()))
        

    def __str__(self):
        cls = self.__class__
        if self._name_ is not None:
            return "%s" % (to_pascal_case(self._name_))
        members, _ = _decompose(cls, self._value_)
        if len(members) == 1 and members[0]._name_ is None:
            return "%r" % (members[0]._value_)
        else:
            return "%s" % (
                ", ".join(
                    [to_pascal_case(str(m._name_ or m._value_)) for m in members]
                ),
            )

    def __repr__(self):
        return str(self)

class MS_PKI_CERTIFICATE_NAME_FLAG(IntFlag):
    NONE = 0
    ENROLLEE_SUPPLIES_SUBJECT = 1
    ADD_EMAIL = 0x00000002
    ADD_OBJ_GUID = 0x00000004
    OLD_CERT_SUPPLIES_SUBJECT_AND_ALT_NAME = 0x00000008
    ADD_DIRECTORY_PATH = 0x00000100
    ENROLLEE_SUPPLIES_SUBJECT_ALT_NAME = 0x00010000
    SUBJECT_ALT_REQUIRE_DOMAIN_DNS = 0x00400000
    SUBJECT_ALT_REQUIRE_SPN = 0x00800000
    SUBJECT_ALT_REQUIRE_DIRECTORY_GUID = 0x01000000
    SUBJECT_ALT_REQUIRE_UPN = 0x02000000
    SUBJECT_ALT_REQUIRE_EMAIL = 0x04000000
    SUBJECT_ALT_REQUIRE_DNS = 0x08000000
    SUBJECT_REQUIRE_DNS_AS_CN = 0x10000000
    SUBJECT_REQUIRE_EMAIL = 0x20000000
    SUBJECT_REQUIRE_COMMON_NAME = 0x40000000
    SUBJECT_REQUIRE_DIRECTORY_PATH = 0x80000000

    def __str__(self):
        return self.name

class MS_PKI_PRIVATE_KEY_FLAG(IntFlag):
    REQUIRE_PRIVATE_KEY_ARCHIVAL = 0x00000001
    EXPORTABLE_KEY = 0x00000010
    STRONG_KEY_PROTECTION_REQUIRED = 0x00000020
    REQUIRE_ALTERNATE_SIGNATURE_ALGORITHM = 0x00000040
    REQUIRE_SAME_KEY_RENEWAL = 0x00000080
    USE_LEGACY_PROVIDER = 0x00000100
    ATTEST_NONE = 0x00000000
    ATTEST_REQUIRED = 0x00002000
    ATTEST_PREFERRED = 0x00001000
    ATTESTATION_WITHOUT_POLICY = 0x00004000
    EK_TRUST_ON_USE = 0x00000200
    EK_VALIDATE_CERT = 0x00000400
    EK_VALIDATE_KEY = 0x00000800
    HELLO_LOGON_KEY = 0x00200000

    def __str__(self):
        cls = self.__class__
        if self._name_ is not None:
            return "%s" % (self._name_)
        members, _ = _decompose(cls, self._value_)
        if len(members) == 1 and members[0]._name_ is None:
            return "%r" % (members[0]._value_)
        else:
            return "%s" % (
                ", ".join(
                    [str(m._name_ or m._value_) for m in members]
                ),
            )

class MS_PKI_ENROLLMENT_FLAG(IntFlag):
    NONE = 0x00000000
    INCLUDE_SYMMETRIC_ALGORITHMS = 0x00000001
    PEND_ALL_REQUESTS = 0x00000002
    PUBLISH_TO_KRA_CONTAINER = 0x00000004
    PUBLISH_TO_DS = 0x00000008
    AUTO_ENROLLMENT_CHECK_USER_DS_CERTIFICATE = 0x00000010
    AUTO_ENROLLMENT = 0x00000020
    CT_FLAG_DOMAIN_AUTHENTICATION_NOT_REQUIRED = 0x80
    PREVIOUS_APPROVAL_VALIDATE_REENROLLMENT = 0x00000040
    USER_INTERACTION_REQUIRED = 0x00000100
    ADD_TEMPLATE_NAME = 0x200
    REMOVE_INVALID_CERTIFICATE_FROM_PERSONAL_STORE = 0x00000400
    ALLOW_ENROLL_ON_BEHALF_OF = 0x00000800
    ADD_OCSP_NOCHECK = 0x00001000
    ENABLE_KEY_REUSE_ON_NT_TOKEN_KEYSET_STORAGE_FULL = 0x00002000
    NOREVOCATIONINFOINISSUEDCERTS = 0x00004000
    INCLUDE_BASIC_CONSTRAINTS_FOR_EE_CERTS = 0x00008000
    ALLOW_PREVIOUS_APPROVAL_KEYBASEDRENEWAL_VALIDATE_REENROLLMENT = 0x00010000
    ISSUANCE_POLICIES_FROM_REQUEST = 0x00020000
    SKIP_AUTO_RENEWAL = 0x00040000
    NO_SECURITY_EXTENSION = 0x00080000

    def __str__(self):
        cls = self.__class__
        if self._name_ is not None:
            return "%s" % (self._name_)
        members, _ = _decompose(cls, self._value_)
        if len(members) == 1 and members[0]._name_ is None:
            return "%r" % (members[0]._value_)
        else:
            return "%s" % (
                ", ".join(
                    [str(m._name_ or m._value_) for m in members]
                ),
            )



import hashlib
import base64
from asn1crypto import x509
from bloodhound.ad.utils import ADUtils

from .bloodhound_object import BloodHoundObject
from bofhound.logger import logger, OBJ_EXTRA_FMT, ColorScheme


class BloodHoundNTAuthStore(BloodHoundObject):

    GUI_PROPERTIES = [
        'domain', 'name', 'distinguishedname', 'domainsid', 'isaclprotected',
        'description', 'whencreated', 'certthumbprints'
    ]

    COMMON_PROPERTIES = [
    ]

    def __init__(self, object):
        super().__init__(object)

        self._entry_type = "NTAuthStore"
        self.IsDeleted = False
        self.ContainedBy = {}
        self.IsACLProtected = False

        self.Properties['certthumbprints'] = []

        if 'objectguid' in object.keys():
            self.ObjectIdentifier = object.get("objectguid").upper()

        if 'distinguishedname' in object.keys():
            domain = ADUtils.ldap2domain(object.get('distinguishedname')).upper()
            self.Properties['domain'] = domain
            self.Properties['distinguishedname'] = object.get('distinguishedname').upper()

            # name relies on domain existing, so it can be appended to the end
            if 'name' in object.keys():
                self.Properties['name'] = f"{object.get('name').upper()}@{domain}"

        if 'description' in object.keys():
            self.Properties['description'] = object.get('description')
        else:
            self.Properties['description'] = None

        if 'cacertificate' in object.keys():
            certificate_b64 = object.get("cacertificate")
            
            certificate_b64_list = certificate_b64.split(", ")
            for cert in certificate_b64_list:
                certificate_byte_array = base64.b64decode(cert)
                thumbprint = hashlib.sha1(certificate_byte_array).hexdigest().upper()
                self.Properties['certthumbprints'].append(thumbprint)
            
        if 'ntsecuritydescriptor' in object.keys():
            self.RawAces = object['ntsecuritydescriptor']
        

    def to_json(self, properties_level):
        self.Properties['isaclprotected'] = self.IsACLProtected
        data = super().to_json(properties_level)

        data["Aces"] = self.Aces
        data["ObjectIdentifier"] = self.ObjectIdentifier
        data["IsDeleted"] = self.IsDeleted
        data["IsACLProtected"] = self.IsACLProtected
        data["ContainedBy"] = self.ContainedBy
        
        if "domainsid" in self.Properties:
            data["DomainSID"] = self.Properties["domainsid"]

        return data


import re
import codecs

from bofhound.logger import logger
from bofhound.parsers.generic_parser import GenericParser


#
# This class will be inherited by other parsers since most if not all are based
#   off the same BOF, wrapped by various C2s. These methods can be overridden
#   by child classes to handle specific parsing requirements
#
class LdapSearchBofParser():
    RESULT_DELIMITER = "-"
    RESULT_BOUNDARY_LENGTH = 20
    _COMPLETE_BOUNDARY_LINE = -1


    def __init__(self):
        pass

    #
    # Legacy, used by test cases for 1 liner
    #   Removed from __main__.py to avoid duplicating file reads and formatting
    #
    @staticmethod
    def parse_file(file):
        return LdapSearchBofParser.parse_data(
            LdapSearchBofParser.prep_file(file)
        )

    
    #
    # Replaces parse_file() usage in __main__.py to avoid duplicate file reads
    #
    @staticmethod
    def prep_file(file):
        with codecs.open(file, 'r', 'utf-8') as f:
            contents = f.read()

        return re.sub(r'\n\n\d{2}\/\d{2} (\d{2}:){2}\d{2} UTC \[output\]\nreceived output:\n', '', contents)

    
    #
    # Meat of the parsing logic 
    #
    @staticmethod
    def parse_data(data):
        parsed_objects = []
        current_object = None
        in_result_region = False
        previous_attr = None

        in_result_region = False

        lines = data.splitlines()
        for line in lines:
            is_boundary_line = LdapSearchBofParser._is_boundary_line(line)

            if (not in_result_region and
                  not is_boundary_line):
                continue

            if (is_boundary_line
                  and is_boundary_line != LdapSearchBofParser._COMPLETE_BOUNDARY_LINE):
                while True:
                    try:
                        next_line = next(lines)[1]
                        remaining_length = LdapSearchBofParser._is_boundary_line(next_line, is_boundary_line)

                        if remaining_length:
                            is_boundary_line = remaining_length
                            if is_boundary_line == LdapSearchBofParser._COMPLETE_BOUNDARY_LINE:
                                break
                    except:
                        # probably ran past the end of the iterable
                        break

            if (is_boundary_line):
                if not in_result_region:
                    in_result_region = True
                elif current_object is not None:
                    # self.store_object(current_object)
                    parsed_objects.append(current_object)
                current_object = {}
                continue
            elif re.match("^(R|r)etr(e|i)(e|i)ved \\d+ results?", line):
                #self.store_object(current_object)
                parsed_objects.append(current_object)
                in_result_region = False
                current_object = None
                continue

            data = line.split(': ')

            try:
                # If we previously encountered a control message, we're probably still in the old property
                if len(data) == 1:
                    if previous_attr is not None:
                        value = current_object[previous_attr] + line
                else:
                    data = line.split(':')
                    attr = data[0].strip().lower()
                    value = ''.join(data[1:]).strip()
                    previous_attr = attr

                current_object[attr] = value

            except Exception as e:
                logger.debug(f'Error - {str(e)}')

        return parsed_objects


    # Returns one of the following integers:
    #    0 - This is not a boundary line
    #   -1 - This is a complete boundary line
    #    n - The remaining characters needed to form a complete boundary line
    @staticmethod
    def _is_boundary_line(line, length=RESULT_BOUNDARY_LENGTH):
        line = line.strip()
        chars = set(line)

        if len(chars) == 1 and chars.pop() == LdapSearchBofParser.RESULT_DELIMITER:
            if len(line) == length:
                return -1
            elif len(line) < length:
                return LdapSearchBofParser.RESULT_BOUNDARY_LENGTH - len(line)

        return 0 # Falsey


    #
    # Get local groups, sessions, etc by feeding data to GenericParser class
    #
    @staticmethod
    def parse_local_objects(data):
        return GenericParser.parse_data(data)


import json
import codecs

from .shared_parsers import __all_generic_parsers__


class GenericParser:

    def __init__(self):
        pass


    @staticmethod
    def parse_file(file, is_outflankc2=False):
         with codecs.open(file, 'r', 'utf-8') as f:
            if is_outflankc2:
                return GenericParser.parse_outflank_file(f.read())
            else:    
                return GenericParser.parse_data(f.read())

    
    @staticmethod
    def parse_outflank_file(contents):
        parsed_objects = []

        for line in contents.splitlines():
            event_json = json.loads(line.split('UTC ', 1)[1])

            # we only care about task_resonse events
            if event_json['event_type'] != 'task_response':
                continue
             
            # within task_response events, we only care about tasks with specific BOF names
            if event_json['task']['name'].lower() not in ['netsession2', 'netloggedon2', 'regsession', 'netlocalgrouplistmembers2']:
                continue
            
            parsed_objects.extend(GenericParser.parse_data(event_json['task']['response']))

        return parsed_objects


    @staticmethod
    def parse_data(contents):
        parsed_objects = []
        current_parser = None
        current_object = {}
        
        lines = contents.splitlines()

        for line in lines:
            # if we have no current parser, check and see if the current line is a start boundary
            if current_parser is None:
                for parser in __all_generic_parsers__:
                    if parser.is_start_boundary_line(line):
                        current_parser = parser
                        break

            # if we do have a current parser, check and see if the current line is an end boundary
            else:
                if current_parser is not None:
                    if current_parser.is_end_boundary_line(line):
                        # we've reached the end of the current object, so store it and reset the parser
                        current_object["ObjectType"] = current_parser.OBJECT_TYPE
                        parsed_objects.append(current_object)
                        current_parser = None
                        current_object = {}
                        continue
               
                # if we have a current parser and the current line is not an end boundary, parse the line
                if current_parser is not None:
                    current_object = current_parser.parse_line(line, current_object)

        return parsed_objects
                
                
                    
        


from bloodhound.ad.utils import ADUtils

from .bloodhound_object import BloodHoundObject
from bofhound.logger import logger, OBJ_EXTRA_FMT, ColorScheme


class BloodHoundGroup(BloodHoundObject):

    GUI_PROPERTIES = [
        'distinguishedname', 'samaccountname', 'objectsid',
        'admincount', 'description', 'whencreated',
        'name', 'domain', 'domainsid'
    ]

    COMMON_PROPERTIES = [
        'member', 'memberof'
    ]

    def __init__(self, object):
        super().__init__(object)

        self._entry_type = "Group"
        self.Members = []
        self.Aces = []
        self.ContainedBy = {}
        self.IsDeleted = False
        self.IsACLProtected = False
        self.MemberDNs = []
        self.MemberOfDNs = []
        self.IsACLProtected = False

        if 'distinguishedname' in object.keys() and 'samaccountname' in object.keys():
            domain = ADUtils.ldap2domain(object.get('distinguishedname')).upper()
            name = f'{object.get("samaccountname")}@{domain}'.upper()
            self.Properties["name"] = name
            self.Properties["domain"] = domain
            logger.debug(f"Reading Group object {ColorScheme.group}{name}[/]", extra=OBJ_EXTRA_FMT)

        if 'objectsid' in object.keys():
            #objectid = BloodHoundObject.get_sid(object.get('objectsid', None), object.get('distinguishedname', None))
            objectid = object.get('objectsid')
            self.ObjectIdentifier = objectid
            self.Properties["domainsid"] = objectid.rsplit('-',1)[0]


        if 'distinguishedname' in object.keys():
            self.Properties["distinguishedname"] = object.get('distinguishedname', None).upper()

        if 'admincount' in object.keys():
            self.Properties["admincount"] = int(object.get('admincount')) == 1 # do not move this lower, it may break imports for users
        else:
            self.Properties["admincount"] = False

        if 'description' in object.keys():
            self.Properties["description"] = object.get('description')

        if 'member' in object.keys():
            self.MemberDNs = [f'CN={dn.upper()}' for dn in object.get('member').split(', CN=')]
            if len(self.MemberDNs) > 0:
                self.MemberDNs[0] = self.MemberDNs[0][3:]

        if 'ntsecuritydescriptor' in object.keys():
            self.RawAces = object['ntsecuritydescriptor']

        if 'memberof' in object.keys():
                self.MemberOfDNs = [f'CN={dn.upper()}' for dn in object.get('memberof').split(', CN=')]
                if len(self.MemberOfDNs) > 0:
                    self.MemberOfDNs[0] = self.MemberOfDNs[0][3:]


    def add_group_member(self, object, object_type):
        member = {
            "ObjectIdentifier": object.ObjectIdentifier,
            "ObjectType": object_type
        }
        self.Members.append(member)


    def to_json(self, properties_level):
        group = super().to_json(properties_level)
        group["ObjectIdentifier"] = self.ObjectIdentifier
        group["ContainedBy"] = self.ContainedBy
        group["Aces"] = self.Aces
        group["Members"] = self.Members
        group["IsDeleted"] = self.IsDeleted
        group["IsACLProtected"] = self.IsACLProtected

        return group



from .bloodhound_domain import BloodHoundDomain
from .bloodhound_computer import BloodHoundComputer
from .bloodhound_user import BloodHoundUser
from .bloodhound_group import BloodHoundGroup
from .bloodhound_object import BloodHoundObject
from .bloodhound_schema import BloodHoundSchema
from .bloodhound_ou import BloodHoundOU
from .bloodhound_container import BloodHoundContainer
from .bloodhound_gpo import BloodHoundGPO
from .bloodhound_enterpriseca import BloodHoundEnterpriseCA
from .bloodhound_rootca import BloodHoundRootCA
from .bloodhound_aiaca import BloodHoundAIACA
from .bloodhound_ntauthstore import BloodHoundNTAuthStore
from .bloodhound_issuancepolicy import BloodHoundIssuancePolicy
from .bloodhound_certtemplate import BloodHoundCertTemplate
from .bloodhound_domaintrust import BloodHoundDomainTrust
from .bloodhound_crossref import BloodHoundCrossRef


from .models import LocalGroupMembership, LocalPrivilegedSession, LocalSession, LocalRegistrySession
from bofhound.parsers.shared_parsers import NetSessionBofParser, NetLoggedOnBofParser, NetLocalGroupBofParser, RegSessionBofParser


class LocalBroker:

    def __init__(self):
        self.privileged_sessions        = set()
        self.sessions                   = set()
        self.local_group_memberships    = set()
        self.registry_sessions          = set()


    # take in known domain sids so we can filter out local accounts 
    # and accounts with unknown domains
    def import_objects(self, objects, known_domain_sids):

        for object in objects:

            if object["ObjectType"] == NetLoggedOnBofParser.OBJECT_TYPE:
                priv_session = LocalPrivilegedSession(object)
                if priv_session.should_import():
                    self.privileged_sessions.add(priv_session)

            elif object["ObjectType"] == NetSessionBofParser.OBJECT_TYPE:
                session = LocalSession(object)
                if session.should_import():
                    self.sessions.add(session)
                    
            elif object["ObjectType"] == NetLocalGroupBofParser.OBJECT_TYPE:
                local_group_membership = LocalGroupMembership(object)
                if local_group_membership.should_import(known_domain_sids):
                    self.local_group_memberships.add(local_group_membership)

            elif object["ObjectType"] == RegSessionBofParser.OBJECT_TYPE:
                registry_session = LocalRegistrySession(object)
                if registry_session.should_import(known_domain_sids):
                    self.registry_sessions.add(registry_session)            
        


from bloodhound.ad.utils import ADUtils
from bloodhound.ad.trusts import ADDomainTrust
from impacket.ldap.ldaptypes import LDAP_SID

from bofhound.logger import logger, OBJ_EXTRA_FMT, ColorScheme
from bofhound.ad.models.bloodhound_object import BloodHoundObject
from bofhound.ad.helpers import TrustType, TrustDirection


class BloodHoundDomainTrust(object):
    
    def __init__(self, object):
        # Property for internal processing
        self.LocalDomainDn = ''

        # Property that holds final dict for domain JSON
        # {
        #     "TargetDomainName": "",
        #     "TargetDomainSid": "",
        #     "IsTransitive": "",
        #     "TrustDirection": "",
        #     "TrustType": "",
        #     "SidFilteringEnabled": ""
        # }
        self.TrustProperties = None

        if 'distinguishedname' in object.keys() and 'trustpartner' in object.keys() and \
            'trustdirection' in object.keys() and 'trusttype' in object.keys() and 'trustattributes' in object.keys() and \
            'securityidentifier' in object.keys():
            
            self.LocalDomainDn = BloodHoundObject.get_domain_component(object.get('distinguishedname')).upper()
            trust_partner = object.get('trustpartner').upper()
            domain = ADUtils.ldap2domain(object.get('distinguishedname')).upper()
            logger.debug(f'Reading trust relationship between {ColorScheme.domain}{domain}[/] and {ColorScheme.domain}{trust_partner}[/]', extra=OBJ_EXTRA_FMT)
            domainsid = LDAP_SID()
            domainsid.fromCanonical(object.get('securityidentifier'))
            trust = ADDomainTrust(trust_partner, int(object.get('trustdirection')), object.get('trusttype'), int(object.get('trustattributes')), domainsid.getData())
            self.TrustProperties = trust.to_output()

            # BHCE now wants trusttype and direction defined as string names instead of int values
            
            self.TrustProperties['TrustDirection'] = TrustDirection(self.TrustProperties['TrustDirection']).name
            self.TrustProperties['TrustType'] = TrustType(self.TrustProperties['TrustType']).name



import sys
import asyncio
import logging
import warnings
import base64
from mythic import mythic
from syncer import sync

from bofhound.parsers import LdapSearchBofParser
from bofhound.logger import logger

# Suppress the specific aiohttp ssl warning
warnings.filterwarnings(
    "ignore",
    message="WARNING: By default, AIOHTTPTransport does not verify ssl certificates.*"
)

RETURN_ATTRIBUTES = [
    "id",
    "display_id",
    "domain",
    "user",
    "host",
    "agent_callback_id"
]

###
# Quick and dirty class to hold Mythic callback information
# and allow print statments from the main logic to still work
class MythicCallback:
    def __init__(self, callback, mythic_instance=None):
        self.callback_id = callback["id"]
        self.display_id = callback["display_id"]
        self.domain = callback["domain"]
        self.user = callback["user"]
        self.host = callback["host"]
        self.uuid = callback["agent_callback_id"]
        self.mythic_instance = mythic_instance
    
    
    def __repr__(self):
        return f"Mythic callback {self.callback_id} [{self.uuid}]"


class MythicParser(LdapSearchBofParser):
    mythic_instance = None


    def __init__(self):
        super().__init__()


    async def connect(self, mythic_server, mythic_token):
        logger.debug("Logging into Mythic...")
        try:
            self.mythic_instance = await mythic.login(
                apitoken=mythic_token,
                server_ip=mythic_server,
                server_port=7443,
                timeout=-1,
                logging_level=logging.CRITICAL,
            )
        except Exception as e:
            logger.error("Error logging into Mythic")
            logger.error(e)
            sys.exit(-1)
        
        logger.debug("Logged into Mythic successfully")


    async def collect_callbacks(self):
        logger.debug("Retrieving callbacks from Mythic...")
        try:
            raw_callbacks = await mythic.get_all_callbacks(
                self.mythic_instance,
                custom_return_attributes=",".join(RETURN_ATTRIBUTES)
            )
        except Exception as e:
            logger.error("Error retrieving callbacks from Mythic")
            logger.error(e)
            sys.exit(-1)

        all_callbacks = []
        for callback in raw_callbacks:
            mythic_callback = MythicCallback(callback, self.mythic_instance)
            all_callbacks.append(mythic_callback)

        logger.debug(f"Retrieved {len(all_callbacks)} callbacks from Mythic")
        return all_callbacks


    ###
    # For mythic, instead of processing individual log "files"
    # we will processes the taskings given to individual callbacks
    @staticmethod
    def prep_file(callback):
        tasks = sync(
            MythicParser.get_tasks(
                callback.mythic_instance,
                callback.display_id
            )
        )
        
        data = ""

        # 
        # TODO: can we iterate over only relevant tasks?
        #
        for task in tasks:
            output = sync(
                MythicParser.get_task_output(
                    callback.mythic_instance,
                    task["display_id"]
                )
            )

            for response in output:
                data += base64.b64decode(response["response_text"]).decode("utf-8")
                data += "\n"
        
        return data
    

    @staticmethod
    async def get_task_output(mythic_instance, task_id):
        try:
            output = await mythic.get_all_task_output_by_id(
                mythic_instance,
                task_id
            )
        except Exception as e:
            logger.warning(f"Error retrieving task output from Mythic for task {task_id}")
            logger.warning(e)

        return output
    

    @staticmethod
    async def get_tasks(mythic_instance, id):
        try:
            tasks = await mythic.get_all_tasks(
                mythic_instance,
                callback_display_id=id,
                #custom_return_attributes=",".join(RETURN_ATTRIBUTES)
            )
        except Exception as e:
            logger.warning(f"Error retrieving tasks from Mythic for callback {id}")
            logger.warning(e)

        return tasks


from bofhound.logger import logger, OBJ_EXTRA_FMT, ColorScheme


class BloodHoundCrossRef(object):

    def __init__(self, object):
        self.netBiosName = None
        self.nCName = None
        self.distinguishedName = None

        if 'netbiosname' in object.keys() and 'ncname' in object.keys() and 'distinguishedname' in object.keys():
            self.netBiosName = object.get('netbiosname')
            self.nCName = object.get('ncname').upper()
            self.distinguishedName = object.get('distinguishedname').upper()
            logger.debug(f"Reading CrossRef object {ColorScheme.schema}{self.distinguishedName}[/]", extra=OBJ_EXTRA_FMT)



import calendar
import hashlib
import base64
from asn1crypto import x509
from datetime import datetime
from bloodhound.enumeration.acls import SecurityDescriptor, ACL, ACCESS_ALLOWED_ACE, ACCESS_MASK, ACE, ACCESS_ALLOWED_OBJECT_ACE, has_extended_right, EXTRIGHTS_GUID_MAPPING, can_write_property, ace_applies
from bloodhound.ad.utils import ADUtils

from bofhound.logger import logger, OBJ_EXTRA_FMT, ColorScheme
from bofhound.ad.models.bloodhound_schema import BloodHoundSchema
from bofhound.ad.helpers import PropertiesLevel

# TODO: Move appropriate actions from this class to a super class of Users/Computers/maybe groups?


class BloodHoundObject():

    GUI_PROPERTIES = [
    ]

    COMMON_PROPERTIES = [
    ]

    NEVER_SHOW_PROPERTIES = [
        'ntsecuritydescriptor', 'serviceprincipalname'
    ]

    def __init__(self, object=None):
        self.ObjectIdentifier = None
        self.Aces = []
        self.RawAces = None
        self.Properties = {}

        if isinstance(object, dict):
            # Ensure all keys are lowercase
            for item in object.keys():
                self.Properties[item.lower()] = object[item]

            self.ObjectIdentifier = BloodHoundObject.get_sid(object.get('objectsid', None), object.get('distinguishedname', None))

            if 'distinguishedname' in object.keys():
                self.Properties["distinguishedname"] = object.get('distinguishedname', None).upper()

            self.__parse_whencreated(object)


    def get_primary_membership(self, object):
        """
        Construct primary membership from RID to SID (BloodHound 3.0 only)
        """
        try:
            primarygroupid = object.get('primarygroupid')
            return '%s-%s' % ('-'.join(object.get('objectsid').split('-')[:-1]), primarygroupid)
        except (TypeError, KeyError, AttributeError):
            # Doesn't have a primarygroupid, means it is probably a Group instead of a user
            return None


    def merge_entry(self, object, base_preference=False):
        """Merge the properties of another BloodHoundObject in with this one.

        Keyword arguments:
        object          -- the new object to merge (required)
        base_preference -- whether or not to prefer the base object. If true, self's properties could be overwritten (default False)
        """
        self_attributes = self.__dict__.keys()
        for attr, value in object.__dict__.items():
            if attr == 'Properties':
                for k, v in getattr(object, attr).items():
                    if not k in getattr(self, attr).keys():
                        getattr(self, attr)[k] = v
                    else:
                        if k == 'distinguishedname':
                            if not getattr(self, attr)[k]:
                                getattr(self, attr)[k] = v
                        if not base_preference:
                            if getattr(object, attr).get(k, None):
                                getattr(self, attr)[k] = v

            elif not attr in self_attributes:
                setattr(self, attr, value)
            else:
                if attr == 'ObjectIdentifier':
                    if not self.ObjectIdentifier:
                        setattr(self, attr, value)

                if not base_preference:
                    if getattr(object, attr):
                        setattr(self, attr, value)
    


    def get_distinguished_name(self):
        try:
            return self.Properties['distinguishedname'].upper()
        except KeyError:
            return None


    def get_property(self, property):
        try:
            return self.Properties[property]
        except KeyError:
            return None


    def to_json(self, properties_level):
        data = {
            "Properties": {}
        }

        match properties_level:
            case PropertiesLevel.Standard:
                for property in self.Properties.keys():
                    if property in self.GUI_PROPERTIES \
                        and property not in self.NEVER_SHOW_PROPERTIES:
                        data["Properties"][property] = self.Properties[property]
            case PropertiesLevel.Member:
                for property in self.Properties.keys():
                    if (property in self.COMMON_PROPERTIES or property in self.GUI_PROPERTIES) \
                        and property not in self.NEVER_SHOW_PROPERTIES:
                        data["Properties"][property] = self.Properties[property]
            case PropertiesLevel.All:
                data["Properties"] = self.Properties

        return data


    def __parse_whencreated(self, object):
        whencreated = object.get('whencreated', 0)
        try:
            if not isinstance(whencreated, int):
                whencreated = calendar.timegm(datetime.strptime(whencreated, "%Y%m%d%H%M%S.0Z").timetuple())
            self.Properties['whencreated'] = whencreated
        except:
            self.Properties['whencreated'] = whencreated


    # used by Domains and OUs
    def add_linked_gpo(self, object, gp_link_options):
        enforced = False
        if gp_link_options == '2':
            enforced = True

        link = {
            "GUID": object.ObjectIdentifier,
            "IsEnforced": False
        }
        self.Links.append(link)


    # used by Domains and OUs
    def add_ou_member(self, object, object_type):
        member = {
            "ObjectIdentifier": object.ObjectIdentifier,
            "ObjectType": object_type
        }
        self.ChildObjects.append(member)


    @staticmethod
    def get_sid(sid, dn=None):
        if sid in ADUtils.WELLKNOWN_SIDS:
            domain = ADUtils.ldap2domain(dn).upper()
            PrincipalSid = f'{domain}-{sid}'
        else:
            PrincipalSid = sid

        return PrincipalSid


    # Should probably move to ADDS?
    @staticmethod
    def get_domain_component(dn):
        dc = ''
        for component in dn.split(','):
            if component.startswith('DC='):
                dc += f"{component},"
        return dc[:-1]


    @staticmethod
    def get_dn(domain):
        components = domain.split('.')
        base = ''
        for comp in components:
            base += f',DC={comp}'
        
        return base[1:]
    
    
    @staticmethod
    def get_cn_from_dn(dn):
        for component in dn.split(',', 1):
            if component.startswith('CN='):
                return component[3:]
    
    #
    # for AIACAs, EnterpriseCAs, and RootCAs
    #
    def parse_cacertificate(self, object):
        certificate_b64 = object.get("cacertificate")
            
        certificate_byte_array = base64.b64decode(certificate_b64)
        
        #
        # thumbprint
        #
        thumbprint = hashlib.sha1(certificate_byte_array).hexdigest().upper()
        self.Properties['certthumbprint'] = thumbprint
        
        #
        # certname
        #
        certificate_byte_array = base64.b64decode(certificate_b64)
        ca_cert = x509.Certificate.load(certificate_byte_array)["tbs_certificate"]
        self.x509Certificate = ca_cert # set for post-processing
        self.Properties['certname'] = ca_cert['subject'].native.get('common_name', thumbprint)
        
        #
        # cert chain
        # not sure that Python libs offer a way to build the chain without access to the issuer cert like it seems SharpHound  does
        # https://github.com/BloodHoundAD/SharpHoundCommon/blob/ea6b097927c5bb795adb8589e9a843293d36ae37/src/CommonLib/Processors/LDAPPropertyProcessor.cs#L772
        # so we will have the build the chain manually in post-processing
        #
        self.Properties['certchain'] = []

        #
        # extensions (hasbasicconstraints, basicconstraintpathlength)
        #
        self.Properties['hasbasicconstraints'] = False
        self.Properties['basicconstraintpathlength'] = 0
        for ext in ca_cert['extensions']:
            if ext['extn_id'].native == 'basic_constraints':
                basic_constraints = ext['extn_value'].parsed
                if basic_constraints['path_len_constraint'].native is not None:
                    self.Properties['hasbasicconstraints'] = True
                    self.Properties['basicconstraintpathlength'] = basic_constraints['path_len_constraint'].native
                break




from enum import Enum

class ParserType(Enum):
    LdapsearchBof   = 'ldapsearch'
    BRC4            = 'BRC4'
    HAVOC           = 'Havoc'
    OUTFLANKC2      = 'OutflankC2'
    MYTHIC          = 'Mythic'


import calendar
from datetime import datetime
from bloodhound.ad.utils import ADUtils, LDAP_SID

from .bloodhound_object import BloodHoundObject
from bofhound.logger import logger, OBJ_EXTRA_FMT, ColorScheme


class BloodHoundComputer(BloodHoundObject):

    GUI_PROPERTIES = [
        'domain', 'name', 'distinguishedname', 'domainsid', 'samaccountname',
        'haslaps', 'isaclprotected', 'description', 'whencreated', 'enabled',
        'unconstraineddelegation', 'trustedtoauth', 'isdc', 'lastlogon', 'lastlogontimestamp',
        'pwdlastset', 'serviceprincipalnames', 'email', 'operatingsystem', 'sidhistory'
    ]

    COMMON_PROPERTIES = [
        'useraccountcontrol', 'dnshostname', 'samaccounttype', 'primarygroupid',
        'msds-allowedtodelegateto', 'operatingsystemservicepack',
        'msds-allowedtoactonbehalfofotheridentity', 'ms-mcs-admpwdexpirationtime',
        'memberof'
    ]

    LOCAL_GROUP_SIDS = {
        "administrators": 544,
        "remote desktop users": 555,
        "remote management users": 580,
        "distributed com users": 562
    }

    def __init__(self, object):
        super().__init__(object)

        self._entry_type = "Computer"
        self.not_collected = {
            "Collected": False,
            "FailureReason": None,
            "Results": []
        }
        self.uac = None
        self.IsACLProtected = False
        self.IsDeleted = False
        self.hostname = object.get('dnshostname', None)
        self.PrimaryGroupSid = self.get_primary_membership(object) # Returns none if non-existent
        self.sessions = None #['not currently supported by bofhound']
        self.AllowedToDelegate = []
        self.MemberOfDNs = []
        self.sessions = []
        self.ContainedBy = {}
        self.privileged_sessions = []
        self.registry_sessions = []
        self.local_group_members = {} # {group_name: [{member_sid, member_type}]}

        if 'dnshostname' in object.keys():
            self.hostname = object.get('dnshostname', None)
            self.Properties['name'] = self.hostname.upper()
            logger.debug(f"Reading Computer object {ColorScheme.computer}{self.Properties['name']}[/]", extra=OBJ_EXTRA_FMT)

        if 'msds-allowedtodelegateto' in object.keys():
            self.AllowedToDelegate = object.get('msds-allowedtodelegateto').split(', ')

        if 'useraccountcontrol' in object.keys():
            self.uac = int(object.get('useraccountcontrol'))
            self.Properties['unconstraineddelegation'] = self.uac & 0x00080000 == 0x00080000
            self.Properties['enabled'] = self.uac & 2 == 0
            self.Properties['trustedtoauth'] = self.uac & 0x01000000 == 0x01000000
            self.Properties['isdc'] = self.uac & 0x2000 == 0x2000

        if 'operatingsystem' in object.keys():
            self.Properties['operatingsystem'] = object.get('operatingsystem', 'Unknown')

        if 'operatingsystemservicepack' in object.keys() and 'operatingsystem' in self.Properties:
            self.Properties['operatingsystem'] += f' {object.get("operatingsystemservicepack")}'

        if 'sidhistory' in object.keys():
            self.Properties['sidhistory'] = [LDAP_SID(bsid).formatCanonical() for bsid in object.get('sidhistory', [])]
        else:
            self.Properties['sidhistory'] = []

        if 'distinguishedname' in object.keys():
            domain = ADUtils.ldap2domain(object.get('distinguishedname')).upper()
            self.Properties['domain'] = domain
            if 'samaccountname' in object.keys() and 'dnshostname' not in object.keys():
                samacctname = object.get("samaccountname")
                if samacctname.endswith("$"):
                    name = f'{samacctname[:-1]}.{domain}'.upper()
                else:
                    name = f'{samacctname}.{domain}'.upper()
                self.Properties["name"] = name
                logger.debug(f"Reading Computer object {ColorScheme.computer}{self.Properties['name']}[/]", extra=OBJ_EXTRA_FMT)

        # TODO: HighValue / AdminCount
        self.Properties['highvalue'] = False

        if 'ms-mcs-admpwdexpirationtime' in object.keys():
            self.Properties['haslaps'] = True
        else:
            self.Properties['haslaps'] = False

        if 'lastlogontimestamp' in object.keys():
            self.Properties['lastlogontimestamp'] = ADUtils.win_timestamp_to_unix(
                int(object.get('lastlogontimestamp'))
            )

        if 'lastlogon' in object.keys():
            self.Properties['lastlogon'] = ADUtils.win_timestamp_to_unix(
                int(object.get('lastlogon'))
            )

        if 'pwdlastset' in object.keys():
            self.Properties['pwdlastset'] = ADUtils.win_timestamp_to_unix(
                int(object.get('pwdlastset'))
            )

        if 'serviceprincipalname' in object.keys():
            self.Properties['serviceprincipalnames'] = object.get('serviceprincipalname').split(', ')

        if 'description' in object.keys():
            self.Properties['description'] = object.get('description')
        
        if 'email' in object.keys():
            self.Properties['email'] = object.get('email')

        if 'samaccounttype' in object.keys():
            self.Properties['samaccounttype'] = object.get('samaccounttype')    

        if 'ntsecuritydescriptor' in object.keys():
            self.RawAces = object['ntsecuritydescriptor']

        if 'memberof' in object.keys():
                self.MemberOfDNs = [f'CN={dn.upper()}' for dn in object.get('memberof').split(', CN=')]
                if len(self.MemberOfDNs) > 0:
                    self.MemberOfDNs[0] = self.MemberOfDNs[0][3:]

        if 'email' in object.keys():
            self.Properties['email'] = object.get('email')
        else:
            self.Properties['email'] = None

        if 'description' in object.keys():
            self.Properties['description'] = object.get('description')
        else:
            self.Properties['description'] = None

    def to_json(self, properties_level):
        self.Properties['msds-allowedtodelegateto'] = self.AllowedToDelegate
        self.Properties['isaclprotected'] = self.IsACLProtected
        data = super().to_json(properties_level)
        data["Sessions"] = self.format_session_json(self.sessions)
        data["PrivilegedSessions"] = self.format_session_json(self.privileged_sessions)
        data["RegistrySessions"] = self.format_session_json(self.registry_sessions)
        data["ObjectIdentifier"] = self.ObjectIdentifier
        data["PrimaryGroupSID"] = self.PrimaryGroupSid
        data["AllowedToDelegate"] = self.AllowedToDelegate
        data["AllowedToAct"] = []
        data["HasSidHistory"] = self.Properties.get("sidhistory", [])
        data["DumpSMSAPassword"] = []
        data["LocalGroups"] = self.format_local_group_json()
        data["UserRights"] = []
        data["Status"] = None
        data["IsDeleted"] = self.IsDeleted
        data["ContainedBy"] = self.ContainedBy
        data["Aces"] = self.Aces
        data["IsACLProtected"] = self.IsACLProtected
        data["IsDC"] = self.Properties["isdc"]

        return data

    
    def format_session_json(self, results):
        if len(results) == 0:
            return self.not_collected

        return {
            "Collected": True,
            "FailureReason": None,
            "Results": results
        }

    
    def format_local_group_json(self):
        if len(self.local_group_members) == 0:
            return []

        data = []

        hostname = self.hostname if self.hostname is not None else self.Properties['name']

        for group_name, members in self.local_group_members.items():
            data.append({
                "ObjectIdentifier": f"{self.ObjectIdentifier}-{BloodHoundComputer.LOCAL_GROUP_SIDS[group_name]}",
                "Name": f"{group_name}@{hostname}".upper(),
                "Results": members,
                "Collected": True,
                "FailureReason": None
            })

        return data

    
    # check if a session host's fully qualified hostname matches 
    # the computer's dnshostname attribute
    def matches_dnshostname(self, session_host_fqdn):
        if self.Properties.get('dnshostname', '').upper() == session_host_fqdn.upper():
            return True
        return False

    
    # check if a session host's hostname matches the computer's samaccountname
    def matches_samaccountname(self, session_hostname):
        if self.Properties['samaccountname'].upper() == session_hostname.upper() + '$':
            return True
        return False


    # add a session to the computer object
    def add_session(self, user_sid, session_type):
        session = {
            "UserSID": user_sid,
            "ComputerSID": self.ObjectIdentifier,
        }

        if session_type == 'privileged':
            self.privileged_sessions.append(session)
        elif session_type == 'registry':
            self.registry_sessions.append(session)
        elif session_type == 'session':
            self.sessions.append(session)


    # add a local group member
    def add_local_group_member(self, member_sid, member_type, group_name):
        member = {
                "ObjectIdentifier": member_sid,
                "ObjectType": member_type
        }

        if group_name.lower() not in self.local_group_members.keys():
            self.local_group_members[group_name.lower()] = [ member ]
        else:
            self.local_group_members[group_name.lower()].append(member)    



import hmac
import hashlib
import base64
import requests
import datetime
import urllib3
from pathlib import Path
from typing import Optional

from bofhound.logger import logger


class BloodHoundUploader:

    def __init__(self, server, token_id, token_key):
        self.base_url = server
        self.token_id = token_id
        self.token_key = token_key
        self.upload_job_id = None
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    def create_upload_job(self):
        try:
            r = self._request(
                method="POST",
                uri="/api/v2/file-upload/start"
            )

            if r.status_code != 201:
                logger.error(f"Failed to create upload job: HTTP {r.status_code} - {BloodHoundUploader.get_error(r)}")
                return None
            
            self.upload_job_id = r.json()["data"]["id"]
            logger.debug(f"BloodHound upload job created with ID: {self.upload_job_id}")
            return True
        
        except Exception as e:
            logger.error(f"Error creating upload job")
            logger.error(e)
            return False


    def upload_file(self, file):
        #
        # File comes in as PurePath object
        #  so create a Path object
        #
        file = Path(file)

        content_type = "application/json"
        if file.name.endswith(".zip"):
            content_type = "application/zip"
        
        try:
            r = self._request(
                method="POST",
                uri=f"/api/v2/file-upload/{self.upload_job_id}",
                body=file.read_bytes(),
                content_type=content_type
            )

            if r.status_code != 202:
                logger.error(f"Failed to upload file {file.name}: HTTP {r.status_code} - {BloodHoundUploader.get_error(r)}")

            logger.debug(f"Uploaded file {file.name} to BloodHound: HTTP {r.status_code}")

        except Exception as e:
            logger.error(f"Error uploading file {file.name}")
            logger.error(e)
            return None

    
    def close_upload_job(self):
        try:
            r = self._request(
                method="POST",
                uri=f"/api/v2/file-upload/{self.upload_job_id}/end"
            )

            if r.status_code != 200:
                logger.error(f"Failed to close upload job: HTTP {r.status_code} - {BloodHoundUploader.get_error(r)}")
                return None
            
            logger.debug(f"Ended BloodHound upload job successfully")

        except Exception as e:
            logger.error(f"Error closing upload job")
            logger.error(e)

    #
    # Sign requests
    #   https://bloodhound.specterops.io/integrations/bloodhound-api/working-with-api#use-your-api-key%2Fid-pair
    #
    def _request(self, method: str, uri: str, body: Optional[bytes] = None, content_type: str = "application/json") -> requests.Response:
        digester = hmac.new(self.token_key.encode(), None, hashlib.sha256)
        digester.update(f'{method}{uri}'.encode())
        digester = hmac.new(digester.digest(), None, hashlib.sha256)

        datetime_formatted = datetime.datetime.now().astimezone().isoformat('T')
        digester.update(datetime_formatted[:13].encode())

        digester = hmac.new(digester.digest(), None, hashlib.sha256)

        if body is not None:
            digester.update(body)

        # Perform the request with the signed and expected headers
        return requests.request(
            method=method,
            url=f"{self.base_url}{uri}",
            headers={
                'User-Agent': 'bhe-python-sdk 0001',
                'Authorization': f'bhesignature {self.token_id}',
                'RequestDate': datetime_formatted,
                'Signature': base64.b64encode(digester.digest()),
                'Content-Type': content_type,
            },
            data=body,
            verify=False
        )

    @staticmethod
    def get_error(r):
        try:
            return r.json()["errors"][0]["message"]
        except Exception as e:
            logger.error(f"Error parsing response: {e}")
            return r.text



import re
import codecs
from datetime import datetime as dt

from bofhound.logger import logger
from bofhound.parsers import LdapSearchBofParser


class Brc4LdapSentinelParser(LdapSearchBofParser):
    # BRC4 LDAP Sentinel currently only queries attributes=["*"] and objectClass 
    # is always the top result. May need to be updated in the future.
    START_BOUNDARY = '[+] objectclass                        :'
    END_BOUNDARY = '+-------------------------------------------------------------------+'

    FORMATTED_TS_ATTRS = ['lastlogontimestamp', 'lastlogon', 'lastlogoff', 'pwdlastset', 'accountexpires']
    ISO_8601_TS_ATTRS = ['dscorepropagationdata', 'whenchanged', 'whencreated']
    BRACKETED_ATTRS = ['objectguid']
    SEMICOLON_DELIMITED_ATTRS = ['serviceprincipalname', 'memberof', 'member', 'objectclass']

    def __init__(self):
        pass #self.objects = []

    #
    # Legacy, used by test cases for 1 liner
    #   Removed from __main__.py to avoid duplicating file reads and formatting
    #
    @staticmethod
    def parse_file(file):
        with codecs.open(file, 'r', 'utf-8', errors='ignore') as f:
            return Brc4LdapSentinelParser.parse_data(f.read())


    #
    # Replaces parse_file() usage in __main__.py to avoid duplicate file reads
    #
    @staticmethod
    def prep_file(file):
        with codecs.open(file, 'r', 'utf-8', errors='ignore') as f:
            return f.read()


    @staticmethod
    def parse_data(contents):
        parsed_objects = []
        current_object = None
        in_result_region = False

        in_result_region = False

        lines = contents.splitlines()
        for line in lines:

            if len(line) == 0:
                continue

            is_start_boundary_line = Brc4LdapSentinelParser._is_start_boundary_line(line)
            is_end_boundary_line = Brc4LdapSentinelParser._is_end_boundary_line(line)

            if not in_result_region and not is_start_boundary_line:
                continue

            if is_start_boundary_line:
                if not in_result_region:
                    in_result_region = True

                current_object = {}

            elif is_end_boundary_line:
                parsed_objects.append(current_object)
                in_result_region = False
                current_object = None
                continue

            data = line.split(': ')

            try:
                data = line.split(':', 1)
                attr = data[0].replace('[+]', '').strip().lower()
                value = data[1].strip()

                # BRc4 formats some timestamps for us that we need to revert to raw values
                if attr in Brc4LdapSentinelParser.FORMATTED_TS_ATTRS:
                    if value.lower() in ['never expires', 'value not set']:
                        continue
                    timestamp_obj = dt.strptime(value, '%m/%d/%Y %I:%M:%S %p')
                    value = int((timestamp_obj - dt(1601, 1, 1)).total_seconds() * 10000000)

                if attr in Brc4LdapSentinelParser.ISO_8601_TS_ATTRS:
                    formatted_ts = []
                    for ts in value.split(';'):
                        timestamp_obj = dt.strptime(ts.strip(), "%m/%d/%Y %I:%M:%S %p")
                        timestamp_str = timestamp_obj.strftime("%Y%m%d%H%M%S.0Z")
                        formatted_ts.append(timestamp_str)
                    value = ', '.join(formatted_ts)
                
                # BRc4 formats some attributes with surroudning {} we need to remove
                if attr in Brc4LdapSentinelParser.BRACKETED_ATTRS:
                    value = value[1:-1]

                # BRc4 delimits some list-esque attributes with semicolons
                # when our BH models expect commas
                if attr in Brc4LdapSentinelParser.SEMICOLON_DELIMITED_ATTRS:
                    value = value.replace('; ', ', ')

                # BRc4 puts the trustDirection attribute within securityidentifier
                if attr == 'securityidentifier' and 'trustdirection' in value.lower():
                    trust_direction = value.lower().split('trustdirection ')[1]
                    current_object['trustdirection'] = trust_direction
                    value = value.split('trustdirection: ')[0]
                    continue

                current_object[attr] = value

            except Exception as e:
                logger.debug(f'Error - {str(e)}')

        return parsed_objects


    @staticmethod
    def _is_start_boundary_line(line):
        # BRc4 seems to always have objectClass camelcased, but we'll use lower() just in case
        if line.lower().startswith(Brc4LdapSentinelParser.START_BOUNDARY):
            return True
        return False


    @staticmethod
    def _is_end_boundary_line(line):
        if line == Brc4LdapSentinelParser.END_BOUNDARY:
            return True
        return False



from .local_groupmembership import LocalGroupMembership
from .local_privilegedsession import LocalPrivilegedSession
from .local_session import LocalSession
from .local_registrysession import LocalRegistrySession


import re
import base64
import datetime
from io import BytesIO
from impacket.uuid import string_to_bin
from bloodhound.ad.utils import ADUtils
from bloodhound.enumeration.acls import SecurityDescriptor, ACL, ACCESS_ALLOWED_ACE, ACCESS_MASK, ACE, ACCESS_ALLOWED_OBJECT_ACE, has_extended_right, EXTRIGHTS_GUID_MAPPING, can_write_property, ace_applies

from bofhound.logger import logger
from bofhound.ad.models import BloodHoundComputer, BloodHoundDomain, BloodHoundGroup, BloodHoundObject, BloodHoundSchema, BloodHoundUser, BloodHoundOU, BloodHoundGPO, BloodHoundEnterpriseCA, BloodHoundAIACA, BloodHoundRootCA, BloodHoundNTAuthStore, BloodHoundIssuancePolicy, BloodHoundCertTemplate, BloodHoundContainer, BloodHoundDomainTrust, BloodHoundCrossRef
from bofhound.logger import OBJ_EXTRA_FMT, ColorScheme
from bofhound import console

#
# Add a GUID for enroll to the bloodhound-python mapping we imported
#
EXTRIGHTS_GUID_MAPPING["Enroll"] = string_to_bin("0e10c968-78fb-11d2-90d4-00c04f79dc55")


class ADDS():

    AT_SCHEMAIDGUID = "schemaidguid"
    AT_SAMACCOUNTTYPE = "samaccounttype"
    AT_DISTINGUISHEDNAME = "distinguishedname"
    AT_MSDS_GROUPMSAMEMBERSHIP = "msds-groupmsamembership"
    AT_OBJECTCLASS = "objectclass"
    AT_OBJECTID = "objectsid"
    AT_NAME = "name"
    AT_COMMONNAME = "cn"
    AT_SAMACCOUNTNAME = "samaccountname"
    AT_ORGUNIT = "ou"
    AT_OBJECTGUID = "objectguid"


    def __init__(self):
        self.sid = None
        self.SID_MAP = {} # {sid: BofHoundModel}
        self.DN_MAP = {} # {dn: BofHoundModel}
        self.DOMAIN_MAP = {} # {dc: ObjectIdentifier}
        self.CROSSREF_MAP = {} # { netBiosName: BofHoundModel }
        self.ObjectTypeGuidMap = {} # { Name : schemaIdGuid }
        self.domains = []
        self.users = []
        self.computers = []
        self.groups = []
        self.ous = []
        self.gpos = []
        self.enterprisecas = []
        self.aiacas = []
        self.rootcas = []
        self.ntauthstores = []
        self.issuancepolicies = []
        self.certtemplates = []
        self.containers = []
        self.schemas = []
        self.trusts = []
        self.trustaccounts = []
        self.unknown_objects = []


    def import_objects(self, objects):
        """Parse a list of dictionaries representing attributes of an AD object
            and add or merge them into appropriate lists of objects in the ADDS instance

        objects: [] of {} containing attributes for AD objects
        """

        for object in objects:
            # check if object is a schema - exception for normally required attributes
            schemaIdGuid = object.get(ADDS.AT_SCHEMAIDGUID, None)
            if schemaIdGuid:
                new_schema = BloodHoundSchema(object)
                if new_schema.SchemaIdGuid is not None:
                    self.schemas.append(new_schema)
                    if new_schema.Name not in self.ObjectTypeGuidMap.keys():
                        self.ObjectTypeGuidMap[new_schema.Name] = new_schema.SchemaIdGuid
                continue
            
            # check if object is a crossRef - exception for normally required attributes
            if 'top, crossRef' in object.get(ADDS.AT_OBJECTCLASS, ''):
                new_crossref = BloodHoundCrossRef(object)
                if new_crossref.netBiosName is not None:
                    if new_crossref.netBiosName not in self.CROSSREF_MAP.keys():
                        self.CROSSREF_MAP[new_crossref.netBiosName] = new_crossref
                continue

            #
            # if samaccounttype comes back as something other
            #  than int, skip the object
            #
            try:
                accountType = int(object.get(ADDS.AT_SAMACCOUNTTYPE, 0))
            except:
                continue
                
            target_list = None

            # objectClass: top, container
            # objectClass: top, container, groupPolicyContainer
            # objectClass: top, organizationalUnit

            dn = object.get(ADDS.AT_DISTINGUISHEDNAME, None)
            sid = object.get(ADDS.AT_OBJECTID, None)
            guid = object.get(ADDS.AT_OBJECTGUID, None)

            # SID and DN are required attributes for bofhound objects
            if dn is None or (sid is None and guid is None):
                self.unknown_objects.append(object)
                continue

            originalObject = self.retrieve_object(dn.upper(), sid)
            bhObject = None

            # Groups
            if accountType in [268435456, 268435457, 536870912, 536870913]:
                bhObject = BloodHoundGroup(object)
                target_list = self.groups

            # Users
            elif object.get(ADDS.AT_MSDS_GROUPMSAMEMBERSHIP, b'') != b'' \
                or accountType in [805306368]:
                bhObject = BloodHoundUser(object)
                target_list = self.users

            # Computers
            elif accountType in [805306369]:
                bhObject = BloodHoundComputer(object)
                target_list = self.computers

            # Trust Accounts
            elif accountType in [805306370]:
                self.trustaccounts.append(object)

            # Other Things :)
            else:
                object_class = object.get(ADDS.AT_OBJECTCLASS, '')
                # if 'top, domain' in object_class or 'top, builtinDomain' in object_class:
                if 'top, domain' in object_class:
                    if 'objectsid' in object.keys():
                        bhObject = BloodHoundDomain(object)
                        self.add_domain(bhObject)
                        target_list = self.domains
                # grab domain trusts
                elif 'trustedDomain' in object_class:
                    bhObject = BloodHoundDomainTrust(object)
                    target_list = self.trusts
                # grab OUs
                elif 'top, organizationalUnit' in object_class:
                    bhObject = BloodHoundOU(object)
                    target_list = self.ous
                elif 'container, groupPolicyContainer' in object_class:
                    bhObject = BloodHoundGPO(object)
                    target_list = self.gpos
                # grab PKIs
                elif 'top, certificationAuthority' in object_class:
                    if 'CN=AIA,' in object.get('distinguishedname'):
                        bhObject = BloodHoundAIACA(object)
                        target_list = self.aiacas
                    elif 'CN=Certification Authorities,' in object.get('distinguishedname') :
                        bhObject = BloodHoundRootCA(object)
                        target_list = self.rootcas
                    elif object.get('distinguishedname').upper().startswith('CN=NTAUTHCERTIFICATES,CN=PUBLIC KEY SERVICES,CN=SERVICES,CN=CONFIGURATION,'):
                        bhObject = BloodHoundNTAuthStore(object)
                        target_list = self.ntauthstores
                elif 'top, msPKI-Enterprise-Oid' in object_class:
                    # only want these if flags property is 2, ref: https://github.com/BloodHoundAD/SharpHoundCommon/blob/ea6b097927c5bb795adb8589e9a843293d36ae37/src/CommonLib/Extensions.cs#L402
                    if 'flags' in object.keys():
                        if object.get('flags') == '2':
                            bhObject = BloodHoundIssuancePolicy(object)
                            target_list = self.issuancepolicies
                elif 'top, pKIEnrollmentService' in object_class:
                    bhObject = BloodHoundEnterpriseCA(object)
                    target_list = self.enterprisecas
                # grab PKI Templates
                elif 'top, pKICertificateTemplate' in object_class:
                    bhObject = BloodHoundCertTemplate(object)
                    target_list = self.certtemplates
                elif 'top, container' in object_class:
                    if not (re.search(r'\{.*\},CN=Policies,CN=System,', object.get('distinguishedname')) or 'CN=Operations,CN=DomainUpdates,CN=System' in object.get('distinguishedname')):
                        bhObject = BloodHoundContainer(object)
                        target_list = self.containers
                # some well known SIDs dont return the accounttype property
                elif object.get(ADDS.AT_NAME) in ADUtils.WELLKNOWN_SIDS:
                    bhObject, target_list =  self._lookup_known_sid(object, object.get(ADDS.AT_NAME))
                elif object.get(ADDS.AT_COMMONNAME) in ADUtils.WELLKNOWN_SIDS:
                    bhObject, target_list =  self._lookup_known_sid(object, object.get(ADDS.AT_COMMONNAME))
                else:
                    self.unknown_objects.append(object)


            if originalObject:
                if bhObject:
                    originalObject.merge_entry(bhObject)
                else:
                    bhObject = BloodHoundObject(object)
                    originalObject.merge_entry(bhObject)
            elif bhObject:
                target_list.append(bhObject)
                if not isinstance(bhObject, BloodHoundDomainTrust): # trusts don't have SIDs
                    self.add_object_to_maps(bhObject)


    def add_object_to_maps(self, object:BloodHoundObject):
        if object.ObjectIdentifier:
            self.SID_MAP[object.ObjectIdentifier] = object

        if ADDS.AT_DISTINGUISHEDNAME in object.Properties.keys():
           self.DN_MAP[object.Properties[ADDS.AT_DISTINGUISHEDNAME]] = object


    def add_domain(self, object:BloodHoundObject):
        if ADDS.AT_DISTINGUISHEDNAME in object.Properties.keys() and object.ObjectIdentifier:
            dn = object.Properties[ADDS.AT_DISTINGUISHEDNAME]
            dc = BloodHoundObject.get_domain_component(dn.upper())
            if dc not in self.DOMAIN_MAP:
                self.DOMAIN_MAP[dc] = object.ObjectIdentifier


    def retrieve_object(self, dn=None, sid=None):
        if dn:
            if dn in self.DN_MAP.keys():
                return self.DN_MAP[dn]

        if sid:
            if sid in self.SID_MAP.keys():
                return self.SID_MAP[sid]

        return None


    def recalculate_sid(self, object:BloodHoundObject):
        if 'distinguishedname' in object.Properties.keys():
            # check for wellknown sid
            if object.ObjectIdentifier in ADUtils.WELLKNOWN_SIDS:
                object.Properties['domainsid'] = self.DOMAIN_MAP.get(
                    BloodHoundObject.get_domain_component(object.Properties['distinguishedname']),
                    f"S-????"
                )
                object.ObjectIdentifier = BloodHoundObject.get_sid(object.ObjectIdentifier, object.Properties['distinguishedname'])


    def build_relation(self, object, sid, relation, acetype='', inherited=False):
        if acetype != '':
            raise ValueError("BH 4.0 incompatible output called")

        PrincipalSid = BloodHoundObject.get_sid(sid, object.Properties["distinguishedname"])

        if sid in self.SID_MAP.keys():
            PrincipalType = self.SID_MAP[sid]._entry_type
        elif sid in ADUtils.WELLKNOWN_SIDS.keys():
            PrincipalType = ADUtils.WELLKNOWN_SIDS[sid][1].title()
        else:
            PrincipalType = "Unknown"

        return {'RightName': relation, 'PrincipalSID': PrincipalSid, 'IsInherited': inherited, 'PrincipalType': PrincipalType }

    def calculate_contained(self, object):
        
        if object._entry_type == "Domain":
            return
       
        dn = object.Properties['distinguishedname']
        start = dn.find(',') + 1
        contained_dn = dn[start:]
        start_contained = contained_dn[0:2]
        type_contained = ""
        id_contained = None
        
        match start_contained:
            case "CN":
                if contained_dn.startswith("CN=BUILTIN"):
                    id_contained = "S-1-5-32"
                    type_contained = "Domain"
                else:
                    for cn in self.containers:
                        if cn.Properties["distinguishedname"] == contained_dn:
                            id_contained = cn.ObjectIdentifier
                            type_contained = "Container"
                    if type_contained == "":
                        for obj in self.unknown_objects:
                            if str(obj.get('distinguishedname')).upper() == contained_dn:
                                id_contained = obj.get("objectguid").upper()
                                match obj.get('objectclass'):
                                    case 'top, NTDSService':
                                        type_contained = "Base"
                                    case 'top, container':
                                        type_contained = "Container"
                                    case 'top, configuration':
                                        type_contained = "Configuration"
            case "OU":
                type_contained = "OU"
                for ou in self.ous:
                    if ou.Properties["distinguishedname"] == contained_dn:
                        id_contained = ou.ObjectIdentifier
            case "DC":
                type_contained = "Domain"
                for domain in self.domains:
                    if domain.Properties["distinguishedname"] == contained_dn:
                        id_contained = domain.ObjectIdentifier
            case _:
                return
        
        if type_contained == "":
            return
        else:
            # 
            # We've identified the containing object, set prop on the contained object
            #
            object.ContainedBy = {"ObjectIdentifier":id_contained, "ObjectType":type_contained}


    def process(self):
        all_objects = self.users + self.groups + self.computers + self.domains + self.ous + self.gpos + self.containers \
                        + self.aiacas + self.rootcas + self.enterprisecas + self.certtemplates + self.issuancepolicies \
                        + self.ntauthstores
        
        total_objects = len(all_objects)

        num_parsed_relations = 0
            
        with console.status(f" [bold] Processed {num_parsed_relations} ACLs", spinner="aesthetic") as status:
            for i, object in enumerate(all_objects):
                self.recalculate_sid(object)
                self.calculate_contained(object)
                self.add_domainsid_prop(object)
                try:
                    num_parsed_relations += self.parse_acl(object)
                    status.update(f" [bold] Processing {num_parsed_relations} ACLs --- {i}/{total_objects} objects parsed")
                except:
                    #
                    # Catch the occasional error parinsing ACLs
                    #
                    continue

        logger.info(f"Parsed {num_parsed_relations} ACL relationships")

        with console.status(" [bold] Creating default users", spinner="aesthetic"):
            self.write_default_users()
        logger.info("Created default users")

        with console.status(" [bold] Creating default groups", spinner="aesthetic"):
            self.write_default_groups()
        logger.info("Created default groups")

        with console.status(" [bold] Resolving group memberships", spinner="aesthetic"):
            self.resolve_group_members()
        logger.info("Resolved group memberships")

        with console.status(" [bold] Resolving delegation relationships", spinner="aesthetic"):
            self.resolve_delegation_targets()
        logger.info("Resolved delegation relationships")

        with console.status(" [bold] Resolving OU memberships", spinner="aesthetic"):
            self.resolve_ou_members()
        logger.info("Resolved OU memberships")

        with console.status(" [bold] Linking GPOs to OUs", spinner="aesthetic"):
            self.link_gpos()
        logger.info("Linked GPOs to OUs")

        if len(self.trusts) > 0:
            with console.status(" [bold] Resolving domain trusts", spinner="aesthetic"):
                self.resolve_domain_trusts()
            logger.info("Resolved domain trusts")

        if len(self.aiacas) > 0 or len(self.enterprisecas) > 0:
            with console.status(" [bold] Building CA certificate chains", spinner="aesthetic"):
                self.build_certificate_chains()
            logger.info("Built CA certificate chains")

        if len(self.enterprisecas) > 0:
            with console.status(" [bold] Resolving enabled templates per CA", spinner="aesthetic"):
                for ca in self.enterprisecas:
                    self.resolve_published_templates(ca)
            logger.info("Resolved enabled templates per CA")


    def get_sid_from_name(self, name):
        for entry in self.SID_MAP:
            if(self.SID_MAP[entry].Properties["name"].lower() == name):
                return (entry, self.SID_MAP[entry]._entry_type)
        return (None,None)


    def resolve_delegation_targets(self):
        for object in self.computers + self.users:
            delegatehosts = object.AllowedToDelegate
            resolved_delegation_list = []
            for host in delegatehosts:
                try:
                    target = host.split('/')[1]
                except IndexError:
                    logger.warning('Invalid delegation target: %s', host)
                    continue
                try:
                    (sid, object_type) = self.get_sid_from_name(target.lower())
                    delegation_entry = {"ObjectIdentifier": sid, "ObjectType": object_type}
                    resolved_delegation_list.append(delegation_entry)
                except KeyError:
                    if '.' in target:
                        resolved_delegation_list.append(target.upper())
            if len(delegatehosts) > 0:
                object.Properties['allowedtodelegate'] = delegatehosts
                object.AllowedToDelegate = resolved_delegation_list


    def write_default_users(self):
        """
        Write built-in users to users.json file
        """

        for domain in self.domains:
            domainsid = domain.ObjectIdentifier
            domainname = domain.Properties.get('name', 'UNKNOWN').upper()
            logger.debug(f"Adding default groups for {ColorScheme.domain}{domainname}[/] ({domainsid})", extra=OBJ_EXTRA_FMT)

            user = BloodHoundUser()
            user.AllowedToDelegate = []
            user.ObjectIdentifier = "%s-S-1-5-20" % domainname
            user.PrimaryGroupSID = None
            user.Properties = {
                "domain": domainname,
                "domainsid": domainsid,
                "name": "NT AUTHORITY@%s" % domainname,
                ADDS.AT_DISTINGUISHEDNAME: f"CN=S-1-5-20,CN=FOREIGNSECURITYPRINCIPALS,{BloodHoundObject.get_dn(domainname)}"
            }
            user.Aces = []
            user.SPNTargets = []
            user.HasSIDHistory = []
            user.IsDeleted = False
            user.IsACLProtected = False

            self.users.append(user)


    def write_default_groups(self):
        """
        Put default groups in the groups.json file
        """

        # Domain controllers
        # TODO: Determine root domains
        # 1. Look for Enterprise Admins
        # 2. If no enterprise admins group, treat all as forest roots
        #rootdomains = [ domain for domain in self.domains ]
        for domain in self.domains:
            domainsid = domain.ObjectIdentifier
            domainname = domain.Properties.get('name', 'UNKNOWN').upper()
            logger.debug(f"Adding default groups for {ColorScheme.domain}{domainname}[/] ({domainsid})", extra=OBJ_EXTRA_FMT)

            domain_controllers = [ computer for computer in self.computers if computer.PrimaryGroupSid.endswith('-516') ]

            group = BloodHoundGroup({})
            group.IsDeleted = False
            group.IsACLProtected = False
            group.ObjectIdentifier = "%s-S-1-5-9" % domainname
            group.Properties = {
                "domain": domainname.upper(),
                "name": "ENTERPRISE DOMAIN CONTROLLERS@%s" % domainname,
                ADDS.AT_DISTINGUISHEDNAME: f"CN=S-1-5-9,CN=FOREIGNSECURITYPRINCIPALS,{BloodHoundObject.get_dn(domainname)}"
            }
            group.Members = []
            group.Aces = []

            for computer in domain_controllers:
                memberdata = {
                    "ObjectIdentifier": computer.ObjectIdentifier,
                    "ObjectType": "Computer"
                }
                group.Members.append(memberdata)
            self.groups.append(group)


            # Everyone
            evgroup = BloodHoundGroup({})

            evgroup.IsDeleted = False
            evgroup.IsACLProtected = False
            evgroup.ObjectIdentifier = "%s-S-1-1-0" % domainname
            evgroup.Properties = {
                "domain": domainname,
                "domainsid": domainsid,
                "name": "EVERYONE@%s" % domainname,
                ADDS.AT_DISTINGUISHEDNAME: f"CN=S-1-5-0,CN=FOREIGNSECURITYPRINCIPALS,{BloodHoundObject.get_dn(domainname)}"
            }
            evgroup.Members = []
            evgroup.Aces = []

            self.groups.append(evgroup)

            # Authenticated users
            augroup = BloodHoundGroup({})

            augroup.IsDeleted = False
            augroup.IsACLProtected = False
            augroup.ObjectIdentifier = "%s-S-1-5-11" % domainname
            # Was this a mistake? augroup.ObjectIdentifier = "S-1-5-11"
            augroup.Properties = {
                    "domain": domainname,
                    "domainsid": domainsid,
                    "name": "AUTHENTICATED USERS@%s" % domainname,
                    ADDS.AT_DISTINGUISHEDNAME: f"CN=S-1-5-11,CN=FOREIGNSECURITYPRINCIPALS,{BloodHoundObject.get_dn(domainname)}"
                }
            augroup.Members = []
            augroup.Aces = []

            self.groups.append(augroup)

            # Interactive
            iugroup = BloodHoundGroup({})

            iugroup.IsDeleted = False
            iugroup.IsACLProtected = False
            iugroup.ObjectIdentifier = "%s-S-1-5-4" % domainname
            iugroup.Properties = {
                    "domain": domainname,
                    "domainsid": domainsid,
                    "name": "INTERACTIVE@%s" % domainname,
                    ADDS.AT_DISTINGUISHEDNAME: f"CN=S-1-5-4,CN=FOREIGNSECURITYPRINCIPALS,{BloodHoundObject.get_dn(domainname)}"
                }
            iugroup.Members = []
            iugroup.Aces = []

            self.groups.append(iugroup)


    def resolve_group_members(self):
        for group in self.groups:
            for user in self.users:
                if self._is_member_of(user, group):
                    group.add_group_member(user, "User")
                    logger.debug(f"Resolved {ColorScheme.user}{user.Properties['name']}[/] as member of {ColorScheme.group}{group.Properties['name']}[/]", extra=OBJ_EXTRA_FMT)


            for computer in self.computers:
                if self._is_member_of(computer, group):
                    group.add_group_member(computer, "Computer")
                    logger.debug(f"Resolved {ColorScheme.computer}{computer.Properties['name']}[/] as member of {ColorScheme.group}{group.Properties['name']}[/]", extra=OBJ_EXTRA_FMT)


            for subgroup in self.groups:
                if self._is_nested_group(subgroup, group):
                    group.add_group_member(subgroup, "Group")
                    logger.debug(f"Resolved {ColorScheme.group}{subgroup.Properties['name']}[/] as nested member of {ColorScheme.group}{group.Properties['name']}[/]", extra=OBJ_EXTRA_FMT)

    
    def resolve_ou_members(self):
        for user in self.users:
            ou = self._resolve_object_ou(user)
            if ou is not None:
                ou.add_ou_member(user, "User")
                logger.debug(f"Identified {ColorScheme.user}{user.Properties['name']}[/] as within OU {ColorScheme.ou}{ou.Properties['name']}[/]", extra=OBJ_EXTRA_FMT)

        for group in self.groups:
            ou = self._resolve_object_ou(group)
            if ou is not None:
                ou.add_ou_member(group, "Group")
                logger.debug(f"Identified {ColorScheme.group}{group.Properties['name']}[/] as within OU {ColorScheme.ou}{ou.Properties['name']}[/]", extra=OBJ_EXTRA_FMT)

        for computer in self.computers:
            ou = self._resolve_object_ou(computer)
            if ou is not None:
                ou.add_ou_member(computer, "Computer")
                logger.debug(f"Identified {ColorScheme.computer}{computer.Properties['name']}[/] as within OU {ColorScheme.ou}{ou.Properties['name']}[/]", extra=OBJ_EXTRA_FMT)

        for nested_ou in self.ous:
            ou = self._resolve_nested_ou(nested_ou)
            if ou is not None:
                ou.add_ou_member(nested_ou, "OU")
                logger.debug(f"Identified {ColorScheme.ou}{nested_ou.Properties['name']}[/] as within OU {ColorScheme.ou}{ou.Properties['name']}[/]", extra=OBJ_EXTRA_FMT)
        
        sorted_ous = sorted(self.ous, key=lambda x: len(x.Properties['distinguishedname']), reverse=True)
        
        for ou in sorted_ous:
            affectedcomputers = []
            affectedusers = []
            for childobject in ou.ChildObjects:
                match childobject["ObjectType"] :
                    case "Computer":
                        affectedcomputers.append(childobject)
                    case "User":
                        affectedusers.append(childobject)
                    case "OU":
                        childid = childobject["ObjectIdentifier"]
                        for childou in sorted_ous:
                            if childou.ObjectIdentifier == childid:
                                affectedcomputers = affectedcomputers + childou.AffectedComputers
                                affectedusers = affectedusers + childou.AffectedUsers

            ou.AffectedComputers = affectedcomputers
            ou.AffectedUsers = affectedusers

        sorted_domains = sorted(self.domains, key=lambda x: len(x.Properties['distinguishedname']), reverse=True)
        
        for domain in sorted_domains:
            affectedcomputers = []
            affectedusers = []
            for childobject in domain.ChildObjects:
                match childobject["ObjectType"] :
                    case "Computer":
                        affectedcomputers.append(childobject)
                    case "User":
                        affectedusers.append(childobject)
                    case "OU":
                        childid = childobject["ObjectIdentifier"]
                        for childou in sorted_ous:
                            if childou.ObjectIdentifier == childid:
                                affectedcomputers = affectedcomputers + childou.AffectedComputers
                                affectedusers = affectedusers + childou.AffectedUsers

            domain.AffectedComputers = affectedcomputers
            domain.AffectedUsers = affectedusers


    def link_gpos(self):
        # BHCE appears to now require domainsid prop on GPOs
        for gpo in self.gpos:
            self.add_domainsid_prop(gpo)

        for object in self.ous + self.domains:
            if object._entry_type == 'OU':
                self.add_domainsid_prop(object) # since OUs don't have a SID to get a domainsid from

            for gplink in object.GPLinks:
                if gplink[0] in self.DN_MAP.keys():
                    gpo = self.DN_MAP[gplink[0]]
                    object.add_linked_gpo(gpo, gplink[1])

                    if object._entry_type == 'Domain':
                       logger.debug(f"Linked {ColorScheme.gpo}{gpo.Properties['name']}[/] to domain {ColorScheme.domain}{object.Properties['name']}[/]", extra=OBJ_EXTRA_FMT)
                    else:
                        logger.debug(f"Linked {ColorScheme.gpo}{gpo.Properties['name']}[/] to OU {ColorScheme.ou}{object.Properties['name']}[/]", extra=OBJ_EXTRA_FMT)
    

    def resolve_domain_trusts(self):
        for trust in self.trusts:
            if trust.TrustProperties is not None:
                # Start by trying to add the target domain's sid if we have it
                target_domain_dn = ADUtils.domain2ldap(trust.TrustProperties['TargetDomainName'])
                if target_domain_dn in self.DOMAIN_MAP.keys():
                    trust.TrustProperties['TargetDomainSid'] = self.DOMAIN_MAP[target_domain_dn]

                # Append the trust dict to the origin domain's trust list
                if trust.LocalDomainDn in self.DOMAIN_MAP.keys():
                    for domain in self.domains:
                        if trust.LocalDomainDn == domain.Properties['distinguishedname']:
                            # don't add trust relationships more than once!
                            if not any(prior['TargetDomainName'] == trust.TrustProperties['TargetDomainName'] for prior in domain.Trusts):
                                domain.Trusts.append(trust.TrustProperties)
                            break


    def add_domainsid_prop(self, object):
        dc = BloodHoundObject.get_domain_component(object.Properties["distinguishedname"])
        if dc in self.DOMAIN_MAP.keys():
            object.Properties["domainsid"] = self.DOMAIN_MAP[dc]


    def resolve_trust_relationships(self):
        pass


    def resolve_published_templates(self, entry:BloodHoundEnterpriseCA):
        if hasattr(entry, 'CertTemplates') and entry.CertTemplates:
            for template_name in entry.CertTemplates :
                for template in self.certtemplates:
                    if template.Properties['name'].split('@')[0].lower() == template_name.lower() \
                    and template.Properties['domain'] == entry.Properties['domain']:
                        entry.EnabledCertTemplates.append({"ObjectIdentifier": template.ObjectIdentifier.upper(), "ObjectType": "CertTemplate"})


    # Returns int: number of relations parsed
    def parse_acl(self, entry:BloodHoundObject):
        if not entry.RawAces:
            return 0

        try:
            value = base64.b64decode(entry.RawAces)
        except:
            logger.warning(f'Error base64 decoding nTSecurityDescriptor attribute on {entry._entry_type} {entry.Properties["name"]}')
            return 0

        if not value:
            return 0
        sd = SecurityDescriptor(BytesIO(value))
        # Check for protected DACL flag
        entry.IsACLProtected = sd.has_control(sd.PD)
        relations = []

        # Parse owner
        osid = str(sd.owner_sid)
        ignoresids = ["S-1-3-0", "S-1-5-18", "S-1-5-10"]
        # Ignore Creator Owner or Local System
        if osid not in ignoresids:
            relations.append(self.build_relation(entry, osid, 'Owns', inherited=False))
        for ace_object in sd.dacl.aces:
            if ace_object.ace.AceType != 0x05 and ace_object.ace.AceType != 0x00:
                # These are the only two aces we care about currently
                #logger.debug('Don\'t care about acetype %d', ace_object.ace.AceType)
                continue
            # Check if sid is ignored
            sid = str(ace_object.acedata.sid)
            # Ignore Creator Owner or Local System
            if sid in ignoresids:
                continue
            if ace_object.ace.AceType == 0x05:
                is_inherited = ace_object.has_flag(ACE.INHERITED_ACE)
                # ACCESS_ALLOWED_OBJECT_ACE
                if not ace_object.has_flag(ACE.INHERITED_ACE) and ace_object.has_flag(ACE.INHERIT_ONLY_ACE):
                    # ACE is set on this object, but only inherited, so not applicable to us
                    continue

                # Check if the ACE has restrictions on object type (inherited case)
                if ace_object.has_flag(ACE.INHERITED_ACE) \
                    and ace_object.acedata.has_flag(ACCESS_ALLOWED_OBJECT_ACE.ACE_INHERITED_OBJECT_TYPE_PRESENT):
                    # Verify if the ACE applies to this object typ
                    try:
                        if not ace_applies(ace_object.acedata.get_inherited_object_type().lower(), entry._entry_type, self.ObjectTypeGuidMap):
                            continue
                    except KeyError:
                        pass
                mask = ace_object.acedata.mask
                # Now the magic, we have to check all the rights BloodHound cares about

                # Check generic access masks first
                if mask.has_priv(ACCESS_MASK.GENERIC_ALL) or mask.has_priv(ACCESS_MASK.WRITE_DACL) \
                    or mask.has_priv(ACCESS_MASK.WRITE_OWNER) or mask.has_priv(ACCESS_MASK.GENERIC_WRITE):
                    # For all generic rights we should check if it applies to our object type

                    try:
                        if ace_object.acedata.has_flag(ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT) \
                            and not ace_applies(ace_object.acedata.get_object_type().lower(), entry._entry_type, self.ObjectTypeGuidMap):
                            # If it does not apply, break out of the loop here in order to
                            # avoid individual rights firing later on
                            continue
                    except KeyError:
                        pass

                    # Check from high to low, ignore lower privs which may also match the bitmask,
                    # even though this shouldn't happen since we check for exact matches currently
                    if mask.has_priv(ACCESS_MASK.GENERIC_ALL):
                        # Report this as LAPS rights if it's a computer object AND laps is enabled

                        if entry._entry_type.lower() == 'computer' and \
                        ace_object.acedata.has_flag(ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT) and \
                        'haslaps' in entry.Properties.keys() and \
                        'ms-mcs-admpwd' in self.ObjectTypeGuidMap.keys():
                            if ace_object.acedata.get_object_type().lower() == self.ObjectTypeGuidMap['ms-mcs-admpwd']:
                                relations.append(self.build_relation(entry, sid, 'ReadLAPSPassword', inherited=is_inherited))
                        else:
                            relations.append(self.build_relation(entry, sid, 'GenericAll', inherited=is_inherited))
                        continue

                    if mask.has_priv(ACCESS_MASK.GENERIC_WRITE):
                        relations.append(self.build_relation(entry, sid, 'GenericWrite', inherited=is_inherited))
                        # Don't skip this if it's the domain object, since BloodHound reports duplicate
                        # rights as well, and this might influence some queries
                        if entry._entry_type.lower() != 'domain' and entry._entry_type.lower() != 'computer':
                            continue

                    # These are specific bitmasks so don't break the loop from here
                    if mask.has_priv(ACCESS_MASK.WRITE_DACL):
                        relations.append(self.build_relation(entry, sid, 'WriteDacl', inherited=is_inherited))

                    if mask.has_priv(ACCESS_MASK.WRITE_OWNER):
                        relations.append(self.build_relation(entry, sid, 'WriteOwner', inherited=is_inherited))

                # Property write privileges
                writeprivs = ace_object.acedata.mask.has_priv(ACCESS_MASK.ADS_RIGHT_DS_WRITE_PROP)
                if writeprivs:
                    # GenericWrite
                    if entry._entry_type.lower() in ['user', 'group', 'computer', 'gpo'] and not ace_object.acedata.has_flag(ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT):
                        relations.append(self.build_relation(entry, sid, 'GenericWrite', inherited=is_inherited))
                    if entry._entry_type.lower() == 'group' and can_write_property(ace_object, EXTRIGHTS_GUID_MAPPING['WriteMember']):
                        relations.append(self.build_relation(entry, sid, 'AddMember', '', inherited=is_inherited))
                    if entry._entry_type.lower() == 'computer' and can_write_property(ace_object, EXTRIGHTS_GUID_MAPPING['AllowedToAct']):
                        relations.append(self.build_relation(entry, sid, 'AddAllowedToAct', '', inherited=is_inherited))
                    # Property set, but ignore Domain Admins since they already have enough privileges anyway
                    if entry._entry_type.lower() == 'computer' and can_write_property(ace_object, EXTRIGHTS_GUID_MAPPING['UserAccountRestrictionsSet']) and not sid.endswith('-512'):
                        relations.append(self.build_relation(entry, sid, 'WriteAccountRestrictions', '', inherited=is_inherited))

                    # Since 4.0
                    # Key credential link property write rights
                    if entry._entry_type.lower() in ['user', 'computer'] and ace_object.acedata.has_flag(ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT) \
                    and 'ms-ds-key-credential-link' in self.ObjectTypeGuidMap and ace_object.acedata.get_object_type().lower() == self.ObjectTypeGuidMap['ms-ds-key-credential-link']:
                        relations.append(self.build_relation(entry, sid, 'AddKeyCredentialLink', inherited=is_inherited))

                    # ServicePrincipalName property write rights (exclude generic rights)
                    if entry._entry_type.lower() == 'user' and ace_object.acedata.has_flag(ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT) \
                    and ace_object.acedata.get_object_type().lower() == 'f3a64788-5306-11d1-a9c5-0000f80367c1':
                        relations.append(self.build_relation(entry, sid, 'WriteSPN', inherited=is_inherited))

                    #
                    # Rights for certificate templates
                    #
                    if entry._entry_type.lower() == 'pki template' and ace_object.acedata.has_flag(ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT) \
                    and ace_object.acedata.get_object_type().lower() == 'ea1dddc4-60ff-416e-8cc0-17cee534bce7':
                        relations.append(self.build_relation(entry, sid, 'WritePKINameFlag', inherited=is_inherited))
                    
                    if entry._entry_type.lower() == 'pki template' and ace_object.acedata.has_flag(ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT) \
                    and ace_object.acedata.get_object_type().lower() == 'd15ef7d8-f226-46db-ae79-b34e560bd12c':
                        relations.append(self.build_relation(entry, sid, 'WritePKIEnrollmentFlag', inherited=is_inherited))

                elif ace_object.acedata.mask.has_priv(ACCESS_MASK.ADS_RIGHT_DS_SELF):
                    # Self add - since 4.0
                    if entry._entry_type.lower() == 'group' and ace_object.acedata.data.ObjectType == EXTRIGHTS_GUID_MAPPING['WriteMember']:
                        relations.append(self.build_relation(entry, sid, 'AddSelf', '', inherited=is_inherited))

                # Property read privileges
                if ace_object.acedata.mask.has_priv(ACCESS_MASK.ADS_RIGHT_DS_READ_PROP):
                    if entry._entry_type.lower() == 'computer' and \
                    ace_object.acedata.has_flag(ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT) and \
                    'haslaps' in entry.Properties.keys() and \
                    'ms-mcs-admpwd' in self.ObjectTypeGuidMap.keys():
                        if ace_object.acedata.get_object_type().lower() == self.ObjectTypeGuidMap['ms-mcs-admpwd']:
                           relations.append(self.build_relation(entry, sid, 'ReadLAPSPassword', inherited=is_inherited))

                # Extended rights
                control_access = ace_object.acedata.mask.has_priv(ACCESS_MASK.ADS_RIGHT_DS_CONTROL_ACCESS)
                if control_access:
                    # All Extended
                    if entry._entry_type.lower() in ['user', 'domain'] and not ace_object.acedata.has_flag(ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT):
                        relations.append(self.build_relation(entry, sid, 'AllExtendedRights', '', inherited=is_inherited))
                    if entry._entry_type.lower() == 'computer' and not ace_object.acedata.has_flag(ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT):
                        relations.append(self.build_relation(entry, sid, 'AllExtendedRights', '', inherited=is_inherited))
                    if entry._entry_type.lower() == 'domain' and has_extended_right(ace_object, EXTRIGHTS_GUID_MAPPING['GetChanges']):
                        relations.append(self.build_relation(entry, sid, 'GetChanges', '', inherited=is_inherited))
                    if entry._entry_type.lower() == 'domain' and has_extended_right(ace_object, EXTRIGHTS_GUID_MAPPING['GetChangesAll']):
                        relations.append(self.build_relation(entry, sid, 'GetChangesAll', '', inherited=is_inherited))
                    if entry._entry_type.lower() == 'domain' and has_extended_right(ace_object, EXTRIGHTS_GUID_MAPPING['GetChangesInFilteredSet']):
                        relations.append(self.build_relation(entry, sid, 'GetChangesInFilteredSet', '', inherited=is_inherited))
                    if entry._entry_type.lower() == 'user' and has_extended_right(ace_object, EXTRIGHTS_GUID_MAPPING['UserForceChangePassword']):
                        relations.append(self.build_relation(entry, sid, 'ForceChangePassword', '', inherited=is_inherited))

                    #
                    # Rights for certificate templates
                    #
                    if entry._entry_type.lower() in ['pki template', 'enterpriseca'] and has_extended_right(ace_object, EXTRIGHTS_GUID_MAPPING['Enroll']):
                        relations.append(self.build_relation(entry, sid, 'Enroll', '', inherited=is_inherited))


            if ace_object.ace.AceType == 0x00:
                is_inherited = ace_object.has_flag(ACE.INHERITED_ACE)
                mask = ace_object.acedata.mask
                # ACCESS_ALLOWED_ACE
                if not ace_object.has_flag(ACE.INHERITED_ACE) and ace_object.has_flag(ACE.INHERIT_ONLY_ACE):
                    # ACE is set on this object, but only inherited, so not applicable to us
                    continue

                if mask.has_priv(ACCESS_MASK.GENERIC_ALL):
                    # Generic all includes all other rights, so skip from here
                    relations.append(self.build_relation(entry, sid, 'GenericAll', inherited=is_inherited))
                    continue

                if mask.has_priv(ACCESS_MASK.ADS_RIGHT_DS_WRITE_PROP):
                    # Genericwrite is only for properties, don't skip after
                    if entry._entry_type.lower() in ['user', 'group', 'computer', 'gpo']:
                        relations.append(self.build_relation(entry, sid, 'GenericWrite', inherited=is_inherited))

                if mask.has_priv(ACCESS_MASK.WRITE_OWNER):
                    relations.append(self.build_relation(entry, sid, 'WriteOwner', inherited=is_inherited))

                # For users and domain, check extended rights
                if entry._entry_type.lower() in ['user', 'domain'] and mask.has_priv(ACCESS_MASK.ADS_RIGHT_DS_CONTROL_ACCESS):
                    relations.append(self.build_relation(entry, sid, 'AllExtendedRights', '', inherited=is_inherited))

                if entry._entry_type.lower() == 'computer' and mask.has_priv(ACCESS_MASK.ADS_RIGHT_DS_CONTROL_ACCESS) and \
                sid != "S-1-5-32-544" and not sid.endswith('-512'):
                    relations.append(self.build_relation(entry, sid, 'AllExtendedRights', '', inherited=is_inherited))

                if mask.has_priv(ACCESS_MASK.WRITE_DACL):
                    relations.append(self.build_relation(entry, sid, 'WriteDacl', inherited=is_inherited))

        entry.Aces = relations

        return len(relations)


    def _is_member_of(self, member, group):
        if ADDS.AT_DISTINGUISHEDNAME in member.Properties.keys():
            if member.Properties["distinguishedname"] in group.MemberDNs:
                return True
            
        # BRc4 does not use DN in groups' member attribute, so we have
        # to check membership from the other side of the relationship
        if ADDS.AT_DISTINGUISHEDNAME in group.Properties.keys():
            if group.Properties["distinguishedname"] in member.MemberOfDNs:
                return True

        if member.PrimaryGroupSid == group.ObjectIdentifier:
            return True

        return False


    def _is_nested_group(self, subgroup, group):
        if ADDS.AT_DISTINGUISHEDNAME in subgroup.Properties.keys():
            if subgroup.Properties["distinguishedname"] in group.MemberDNs:
                return True

        if ADDS.AT_DISTINGUISHEDNAME in group.Properties.keys():    
            # BRc4 does not use DN in groups' member attribute, so we have
            # to check membership from the other side of the relationship
            if group.Properties["distinguishedname"] in subgroup.MemberOfDNs:
                return True
            
        return False


    def _resolve_object_ou(self, object):
        if "OU=" in object.Properties["distinguishedname"]:
            target_ou = "OU=" + object.Properties["distinguishedname"].split("OU=", 1)[1]
            for ou in self.ous:
                if ou.Properties["distinguishedname"] == target_ou:
                    return ou
        return None

    
    def _resolve_nested_ou(self, nested_ou):
        dn = nested_ou.Properties["distinguishedname"]
        # else is top-level OU
        if len(dn.split("OU=")) > 2:
            target_ou = "OU=" + dn.split("OU=", 2)[2]
            for ou in self.ous:
                if ou.Properties["distinguishedname"] == target_ou:
                    return ou
        else:
            dc = BloodHoundObject.get_domain_component(dn)
            for domain in self.domains:
                if dc == domain.Properties[self.AT_DISTINGUISHEDNAME]:
                    return domain
        return None


    def _lookup_known_sid(self, object, sid):
        known_sid_type = ADUtils.WELLKNOWN_SIDS[sid][1]
        if known_sid_type == "USER":
            bhObject = BloodHoundUser(object)
            target_list = self.users
        elif known_sid_type == "COMPUTER":
            bhObject = BloodHoundComputer(object)
            target_list = self.computers
        elif known_sid_type == "GROUP":
            bhObject = BloodHoundGroup(object)
            target_list = self.groups
        bhObject.Properties["name"] = ADUtils.WELLKNOWN_SIDS[sid][0].upper()
        return bhObject, target_list
    

    def _get_domain_sid_from_netbios_name(self, nbtns_domain):
        if nbtns_domain in self.CROSSREF_MAP.keys():
            dn = self.CROSSREF_MAP[nbtns_domain].distinguishedName
            if dn in self.DOMAIN_MAP.keys():
                return self.DOMAIN_MAP[dn]
        return None


    # process local group memberships and sessions
    def process_local_objects(self, broker):
        for computer in self.computers:
            self.process_privileged_sessions(broker.privileged_sessions, computer)
            self.process_registry_sessions(broker.registry_sessions, computer)
            self.process_sessions(broker.sessions, computer)
            self.process_local_group_memberships(broker.local_group_memberships, computer)


        if len(broker.local_group_memberships) > 0:
            logger.info(f"Resolved local group memberships")

        if len(broker.privileged_sessions) > 0 \
            or len(broker.registry_sessions) > 0 \
            or len(broker.sessions) > 0:

            logger.info(f"Resolved sessions")


    # correlate privileged sessions to BH Computer objects
    def process_privileged_sessions(self, privileged_sessions, computer_object):
        for session in privileged_sessions:
            # skip sessions that have already been matched to a computer object
            if session.matched:
                continue
            
            computer_found = False

            # first we'll try to directly match the session host's dns name to a
            # computer object's dNSHostName attribute
            if session.host_fqdn is not None:
                if computer_object.matches_dnshostname(session.host_fqdn):
                    computer_found = True

            # second we'll check to see if the host's DNS domain is a known domain
            # converting the host DNS suffix to a domain component could be problematic?
            if session.host_domain is not None and not computer_found:
                dc = BloodHoundObject.get_dn(session.host_domain.upper())
                domain_sid = self.DOMAIN_MAP.get(dc, None)

                # if we have the domain, check for a computer with samaccountname host$
                # and the domain's sid 
                if domain_sid is not None:
                    if computer_object.matches_samaccountname(session.host_name) and \
                        computer_object.ObjectIdentifier.startswith(domain_sid):

                        computer_found = True

            # if we've got the computer, then try to find the user's SID
            if not computer_found:
                continue

            match_users = [user for user in self.users if user.Properties.get('samaccountname', '').lower() == session.user.lower()]
            if len(match_users) > 1:
                logger.warning(f"Multiple users with sAMAccountName {ColorScheme.user}{session.user}[/] found for privileged session")
                # TODO: implement NetBIOS domain name handling
                continue
            elif len(match_users) == 1:
                user_sid = match_users[0].ObjectIdentifier
                computer_object.add_session(user_sid, "privileged")
                logger.debug(f"Resolved privileged session on {ColorScheme.computer}{computer_object.Properties['name']}[/]", extra=OBJ_EXTRA_FMT)

    
    # correlate registry sessions to BH Computer objects
    def process_registry_sessions(self, registry_sessions, computer_object):
        for session in registry_sessions:
            # skip sessions that have already been matched to a computer object
            if session.matched:
                continue
            
            # first we'll try to directly match the session host's dns name to a
            # computer object's dNSHostName attribute
            if session.host_fqdn is not None:
                if computer_object.matches_dnshostname(session.host_name):
                    session.matched = True
                    computer_object.add_session(session.user_sid, "registry")
                    logger.debug(f"Resolved registry session on {ColorScheme.computer}{computer_object.Properties['name']}[/] via dNSHostName match", extra=OBJ_EXTRA_FMT)
                    continue

            # second we'll check to see if the host's DNS domain is a known domain
            # converting the host DNS suffix to a domain component could be problematic?
            if session.host_domain is not None:
                dc = BloodHoundObject.get_dn(session.host_domain.upper())
                domain_sid = self.DOMAIN_MAP.get(dc, None)

                # if we have the domain, check for a computer with samaccountname host$
                # and the domain's sid 
                if domain_sid is not None:
                    if computer_object.matches_samaccountname(session.host_name) and \
                        computer_object.ObjectIdentifier.startswith(domain_sid):

                        session.matched = True
                        computer_object.add_session(session.user_sid, "registry")
                        logger.debug(f"Resolved registry session on {ColorScheme.computer}{computer_object.Properties['name']}[/] via domain + sAMAccountName match", extra=OBJ_EXTRA_FMT)
                        continue

            # if we don't have the host domain/FQDN from the session, we just try to match samaccountname
            # this is probably only error prone if there multiple domains with the same hostname
            elif computer_object.matches_samaccountname(session.host_name):
                session.matched = True
                computer_object.add_session(session.user_sid, "registry")
                logger.debug(f"Resolved registry session on {ColorScheme.computer}{computer_object.Properties['name']}[/] via fuzzy sAMAccountName match", extra=OBJ_EXTRA_FMT)


    # correlate sessions to BH Computer objects
    def process_sessions(self, sessions, computer_object):
        for session in sessions:
            # skip sessions that have already been matched to a computer object
            if session.matched:
                continue
            
            computer_found = False

            # case 1: we have the host's DNS name
            if session.ptr_record is not None:

                # first try to match dNSHostName
                if computer_object.matches_dnshostname(session.ptr_record):
                    computer_found = True

                # if that doesn't work, try to match the host's domain
                if session.computer_domain is not None and not computer_found:
                    dc = BloodHoundObject.get_dn(session.computer_domain.upper())
                    domain_sid = self.DOMAIN_MAP.get(dc, None)

                    # if we have the domain, check for a computer with samaccountname host$
                    # and the domain's sid 
                    if domain_sid is not None:
                        if computer_object.matches_samaccountname(session.computer_name) and \
                            computer_object.ObjectIdentifier.startswith(domain_sid):

                            computer_found = True

            # case 2: we have the NETBIOS host and domain name
            elif session.computer_netbios_domain is not None:
                domain_sid = self._get_domain_sid_from_netbios_name(session.computer_netbios_domain) 
                if domain_sid is not None:
                    if computer_object.matches_samaccountname(session.computer_name) and computer_object.ObjectIdentifier.startswith(domain_sid):
                        computer_found = True
            
            # if we've got the computer, then try to find the user's SID
            if not computer_found:
                continue
            
            match_users = [user for user in self.users if user.Properties.get('samaccountname', '').lower() == session.username.lower()]
            if len(match_users) > 1:
                logger.warning(f"Multiple users with sAMAccountName {ColorScheme.user}{session.user}[/] found for session")
                # TODO: implement NetBIOS domain name handling
                continue
            elif len(match_users) == 1:
                user_sid = match_users[0].ObjectIdentifier
                computer_object.add_session(user_sid, "session")
                logger.debug(f"Resolved session on {ColorScheme.computer}{computer_object.Properties['name']}[/]", extra=OBJ_EXTRA_FMT)


    # correlate local group memberships to BH Computer objects
    def process_local_group_memberships(self, local_group_memberships, computer_object):
        for member in local_group_memberships:
            # skip memberships that have already been matched to a computer object
            if member.matched:
                continue
            
            computer_found = False

            # first we'll try to directly match the session host's dns name to a
            # computer object's dNSHostName attribute
            if member.host_fqdn is not None:
                if computer_object.matches_dnshostname(member.host_fqdn):
                    computer_found = True

            
            # second we'll check to see if the host's DNS domain is a known domain
            if member.host_domain is not None and not computer_found:
                dc = BloodHoundObject.get_dn(member.host_domain.upper())
                domain_sid = self.DOMAIN_MAP.get(dc, None)

                # if we have the domain, check for a computer with samaccountname host$
                # and the domain's sid 
                if domain_sid is not None:
                    if computer_object.matches_samaccountname(member.host_name) and \
                        computer_object.ObjectIdentifier.startswith(domain_sid):

                        computer_found = True

            # if we've got the computer, then check the sid before submitting
            if not computer_found:
                continue

            color = ColorScheme.user if member.member_sid_type == "User" else ColorScheme.group
                
            computer_object.add_local_group_member(member.member_sid, member.member_sid_type, member.group)
            logger.debug(f"Resolved {color}{member.member}[/] as member of {ColorScheme.group}{member.group}[/] on {ColorScheme.computer}{computer_object.Properties['name']}[/]", extra=OBJ_EXTRA_FMT)


    @staticmethod
    def find_issuer_ca(start_ca_obj, all_ca_obj):
        for potential_issuer in all_ca_obj:
            if start_ca_obj.x509Certificate['issuer'] == potential_issuer.x509Certificate['subject']:
                return potential_issuer
        return None
    

    @staticmethod
    def build_certificate_chain(start_ca_obj, all_ca_obj):
        chain = [start_ca_obj]
        current_ca = start_ca_obj

        while True:
            if current_ca.x509Certificate['subject'] == current_ca.x509Certificate['issuer']:
                # Found a self-signed certificate (root CA)
                break
            
            issuer_ca = ADDS.find_issuer_ca(start_ca_obj, all_ca_obj)
            if not issuer_ca:
                break
            chain.append(issuer_ca)

            if issuer_ca == current_ca:
                # Found a circular reference (potentially a stopgap solution)
                break
            current_ca = issuer_ca

        return [cert.Properties['certthumbprint'] for cert in chain]

    def build_certificate_chains(self):
        for enterpriseca in self.enterprisecas:
            enterpriseca.Properties['certchain'] = ADDS.build_certificate_chain(enterpriseca, self.enterprisecas)
        
        for aiaca in self.aiacas:
            aiaca.Properties['certchain'] = ADDS.build_certificate_chain(aiaca, self.aiacas)
            


import ipaddress

from bofhound.logger import logger, ColorScheme, OBJ_EXTRA_FMT


class LocalPrivilegedSession:
    PS_HOST     = "Host"
    PS_USERNAME = "Username"
    PS_DOMAIN   = "Domain"

    def __init__(self, object):
        self.host_name      = None
        self.host_domain    = None
        self.host_fqdn      = None
        self.user           = None
        self.user_domain    = None

        # will be set to True if correlated to a computer object
        self.matched = False

        try:
            ipaddress.ip_address(object[LocalPrivilegedSession.PS_HOST])
            logger.debug(f"Skipping session on {object[LocalPrivilegedSession.PS_HOST]} due to IP instead of hostname")
            return
        except:
            pass
        
        if LocalPrivilegedSession.PS_USERNAME in object.keys():
            self.user = object[LocalPrivilegedSession.PS_USERNAME]

        if LocalPrivilegedSession.PS_DOMAIN in object.keys():
            self.user_domain = object[LocalPrivilegedSession.PS_DOMAIN]

        if LocalPrivilegedSession.PS_HOST in object.keys():
            self.host_fqdn = object[LocalPrivilegedSession.PS_HOST]
            if '.' in self.host_fqdn:
                self.host_name = self.host_fqdn.split('.')[0]
                self.host_domain = '.'.join(self.host_fqdn.split('.')[1:])
            else:
                logger.debug(f"FQDN missing from hostname for {ColorScheme.user}{self.user}[/] session on {ColorScheme.computer}{self.host_fqdn}[/]", extra=OBJ_EXTRA_FMT)

    def should_import(self):
        # missing required attributes
        if self.host_name is None or self.host_domain is None \
            or self.user is None or self.user_domain is None:
            return False
        
        # do not import computer accounts
        if self.user.endswith('$'):
            return False
        
        # do not import local accounts
        if self.host_name.lower() == self.user_domain.lower():
            return False
        
        logger.debug(f"NetWkstaUserEnum session found for {ColorScheme.user}{self.user}@{self.user_domain}[/] on {ColorScheme.computer}{self.host_fqdn}[/]", extra=OBJ_EXTRA_FMT)
        return True


    # so that a set can be used to keep a unique list of objects
    def __eq__(self, other):
        return (self.host_name, self.host_domain, self.host_fqdn, self.user, self.user_domain) == \
               (other.host_name, other.host_domain, other.host_fqdn, other.user, other.user_domain)


    # so that a set can be used to keep a unique list of objects
    def __hash__(self):
        return hash((self.host_name, self.host_domain, self.host_fqdn, self.user, self.user_domain))


    # for debugging
    def __repr__(self):
        return f"LocalPrivilegedSession(host_name={self.host_name}, host_domain={self.host_domain}, host_fqdn={self.host_fqdn}, user={self.user}, user_domain={self.user_domain})"